import logging

import numpy as np
import pint
from ophyd_async.core import Device

from .motor import Motor

log = logging.getLogger(__name__)

ureg = pint.UnitRegistry()


class AerotechMotor(Motor):
    """Allow an Aerotech stage to fly-scan via the Ophyd FlyerInterface.

    Set *start_position*, *end_position*, and *step_size* in units of
    the motor record (.EGU), and *dwell_time* in seconds. Then the
    remaining components will be calculated accordingly.

    All position or distance components are assumed to be in motor
    record engineering units, unless preceded with "encoder_", in
    which case they are in units of encoder pulses based on the
    encoder resolution.

    The following diagram describes how the various components relate
    to each other. Distances are not to scale::

                 ┌─ encoder_window_start         ┌─ encoder_window_stop
                 │                               │
                 │ |┄┄┄| *step_size*             │
                 │ │   │ encoder_step_size       │
                 │ │   │                         │
      Window:    ├┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┤

      Pulses: ┄┄┄┄┄╨───╨───╨───╨───╨───╨───╨───╨┄┄┄┄┄
               │   │ │                       │ │   └─ taxi_end
               │   │ └─ *start_position*     │ └─ pso_end
               │   └─ pso_start              └─ *end position*
               └─ taxi_start

    Parameters
    ==========
    axis
      The label used by the aerotech controller to refer to this
      axis. Examples include "@0" or "X".
    encoder
      The number of the encoder to track when fly-scanning with this
      device.

    Components
    ==========
    start_position
      User-requested center of the first scan pixel.
    end_position
      User-requested center of the last scan pixel. This is not
      guaranteed and may be adjusted to match the encoder resolution
      of the stage.
    step_size
      How much space desired between points. This is not guaranteed
      and may be adjusted to match the encoder resolution of the
      stage.
    dwell_time
      How long to take, in seconds, moving between points.
    slew_speed
      How fast to move the stage. Calculated from the remaining
      components.
    taxi_start
      The starting point for motor movement during flying, accounts
      for needed acceleration of the motor.
    taxi_end
      The target point for motor movement during flying, accounts
      for needed acceleration of the motor.
    pso_start
      The motor position corresponding to the first PSO pulse.
    pso_end
      The motor position corresponding to the last PSO pulse.
    encoder_step_size
      The number of encoder counts for each pixel.
    encoder_window_start
      The start of the window within which PSO pulses may be emitted,
      in encoder counts. Should be slightly wider than the actual PSO
      range.
    encoder_window_end
      The end of the window within which PSO pulses may be emitted,
      in encoder counts. Should be slightly wider than the actual PSO
      range.

    """

    axis: str
    pixel_positions: np.ndarray = None
    # Internal encoder in the Ensemble to track for flying
    encoder: int
    encoder_direction: int = 1
    encoder_window_min: int = -8388607
    encoder_window_max: int = 8388607

    # Calculated fly-scan signals
    # pso_start = Cpt(Signal, kind=Kind.config)
    # pso_end = Cpt(Signal, kind=Kind.config)
    # encoder_step_size = Cpt(Signal, kind=Kind.config)
    # encoder_window_start = Cpt(Signal, kind=Kind.config)
    # encoder_window_end = Cpt(Signal, kind=Kind.config)
    # disable_window = Cpt(Signal, value=False, kind=Kind.config)
    # encoder_use_window = Cpt(Signal, value=False, kind=Kind.config)

    # # Status signals
    # flying_complete = Cpt(Signal, kind=Kind.omitted)
    # ready_to_fly = Cpt(Signal, kind=Kind.omitted)

    def __init__(
        self, *args, axis: str | None = None, encoder: int | None = None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        # Set up extra calculations for the flyer
        # self.encoder_resolution.subscribe(self._update_fly_params)
        # self.disable_window.subscribe(self._update_fly_params)
        # Save needed axis/encoder values
        self.axis = axis
        self.encoder = encoder

    # def kickoff(self):
    #     """Start a flyer

    #     The status object return is marked as done once flying
    #     has started.

    #     Returns
    #     -------
    #     kickoff_status : StatusBase
    #         Indicate when flying has started.

    #     """

    #     def flight_check(*args, old_value, value, **kwargs) -> bool:
    #         return not bool(old_value) and bool(value)

    #     # Status object is complete when flying has started
    #     self.ready_to_fly.set(False).wait()
    #     status = SubscriptionStatus(self.ready_to_fly, flight_check)
    #     # Taxi the motor
    #     th = threading.Thread(target=self.taxi)
    #     th.daemon = True
    #     th.start()
    #     # Record time of fly start of scan
    #     self.starttime = time.time()
    #     self._taxi_thread = th  # Prevents garbage collection
    #     return status

    # def complete(self):
    #     """Wait for flying to be complete.

    #     This can either be a question ("are you done yet") or a
    #     command ("please wrap up") to accommodate flyers that have a
    #     fixed trajectory (ex. high-speed raster scans) or that are
    #     passive collectors (ex MAIA or a hardware buffer).

    #     In either case, the returned status object should indicate when
    #     the device is actually finished flying.

    #     Returns
    #     -------
    #     complete_status : StatusBase
    #         Indicate when flying has completed
    #     """

    #     # Prepare a callback to check when the motor has stopped moving
    #     def check_flying(*args, old_value, value, **kwargs) -> bool:
    #         "Check if flying is complete."
    #         return bool(value)

    #     # Status object is complete when flying has started
    #     self.flying_complete.set(False).wait()
    #     status = SubscriptionStatus(self.flying_complete, check_flying)
    #     # Iniate the fly scan
    #     th = threading.Thread(target=self.fly)
    #     th.start()
    #     self._fly_thread = th  # Prevents garbage collection
    #     return status

    # def collect(self) -> Generator[Dict, None, None]:
    #     """Retrieve data from the flyer as proto-events
    #     Yields
    #     ------
    #     event_data : dict
    #         Must have the keys {'time', 'timestamps', 'data'}.

    #     """
    #     # np array of pixel location
    #     pixels = self.pixel_positions
    #     # time of scans start taken at Kickoff
    #     starttime = self.starttime
    #     # time of scans at movement stop
    #     endtime = self.endtime
    #     # grab necessary for calculation
    #     accel_time = self.acceleration.get()
    #     dwell_time = self.flyer_dwell_time.get()
    #     step_size = self.flyer_step_size()
    #     slew_speed = step_size / dwell_time
    #     motor_accel = slew_speed / accel_time
    #     # Calculate the time it takes for taxi to reach first pixel
    #     extrataxi = ((0.5 * ((slew_speed**2) / (2 * motor_accel))) / slew_speed) + (
    #         dwell_time / 2
    #     )
    #     taxi_time = accel_time + extrataxi
    #     # Create np array of times for each pixel in seconds since epoch
    #     startpixeltime = starttime + taxi_time
    #     endpixeltime = endtime - taxi_time
    #     scan_time = endpixeltime - startpixeltime
    #     timestamps1 = np.linspace(
    #         startpixeltime, startpixeltime + scan_time, num=len(pixels)
    #     )
    #     timestamps = [round(ts, 8) for ts in timestamps1]
    #     for value, ts in zip(pixels, timestamps):
    #         yield {
    #             "data": {self.name: value, self.user_setpoint.name: value},
    #             "timestamps": {self.name: ts, self.user_setpoint.name: ts},
    #             "time": ts,
    #         }

    # def describe_collect(self):
    #     """Describe details for the collect() method"""
    #     desc = OrderedDict()
    #     desc.update(self.describe())
    #     return {"positions": desc}

    # def fly(self):
    #     # Start the trajectory
    #     destination = self.flyer_taxi_end.get()
    #     log.debug(f"Flying to {destination}.")
    #     self.move(destination, wait=True)
    #     # Wait for the landing
    #     self.disable_pso()
    #     self.flying_complete.set(True).wait()
    #     # Record end time of flight
    #     self.endtime = time.time()

    # def taxi(self):
    #     # import pdb; pdb.set_trace()
    #     self.disable_pso()
    #     # Initalize the PSO
    #     # Move motor to the scan start point
    #     self.move(self.pso_start.get(), wait=True)
    #     # Arm the PSO
    #     self.enable_pso()
    #     self.arm_pso()
    #     # Move the motor to the taxi position
    #     taxi_start = self.flyer_taxi_start.get()
    #     log.debug(f"Taxiing to {taxi_start}.")
    #     self.move(taxi_start, wait=True)
    #     # Set the speed on the motor
    #     self.velocity.set(self.flyer_slew_speed.get()).wait()
    #     # Set timing on the delay for triggering detectors, etc
    #     self.parent.delay.channel_C.delay.put(0)
    #     self.parent.delay.output_CD.polarity.put(self.parent.delay.polarities.NEGATIVE)
    #     # Count-down timer
    #     # for i in range(10, 0, -1):
    #     #     print(f"{i}...", end="", flush=True)
    #     #     time.sleep(1)
    #     # print("Go!")
    #     self.ready_to_fly.set(True)

    # def stage(self, *args, **kwargs):
    #     self.ready_to_fly.set(False).wait()
    #     self.flying_complete.set(False).wait()
    #     self.starttime = None
    #     self.endtime = None
    #     # Save old veolcity to be restored after flying
    #     self.old_velocity = self.velocity.get()
    #     super().stage(*args, **kwargs)

    # def unstage(self, *args, **kwargs):
    #     self.velocity.set(self.old_velocity).wait()
    #     return super().unstage(*args, **kwargs)

    # def move(self, position, wait=True, *args, **kwargs):
    #     motor_status = super().move(position, wait=wait, *args, **kwargs)

    #     def check_readback(*args, old_value, value, **kwargs) -> bool:
    #         "Check if taxiing is complete and flying has begun."
    #         has_arrived = np.isclose(value, position, atol=0.001)
    #         log.debug(
    #             f"Checking readback: {value=}, target: {position}, {has_arrived=}"
    #         )
    #         return has_arrived

    #     # Status object is complete motor reaches target value
    #     readback_status = SubscriptionStatus(self.user_readback, check_readback)
    #     # Prepare the combined status object
    #     status = motor_status & readback_status
    #     if wait:
    #         status.wait()
    #     return readback_status

    # @property
    # def motor_egu_pint(self):
    #     egu = ureg(self.motor_egu.get())
    #     return egu

    # def flyer_step_size(self):
    #     """Calculate the size of each step in a fly scan."""
    #     start_position = self.flyer_start_position.get()
    #     end_position = self.flyer_end_position.get()
    #     num_points = self.flyer_num_points.get()
    #     step_size = abs(start_position - end_position) / (num_points - 1)
    #     return step_size

    # def _update_fly_params(self, *args, **kwargs):
    #     """Calculate new fly-scan parameters based on signal values.

    #     Implementation courtesy of Alan Kastengren.

    #     Computes several parameters describing the fly scan motion.
    #     These include the actual start position of the motor, the
    #     actual distance (in encoder counts and distance) between PSO
    #     pulses, the end position of the motor, and where PSO pulses
    #     are expected to occcur.  This code ensures that each PSO delta
    #     is an integer number of encoder counts, since this is how the
    #     PSO triggering works in hardware.

    #     These calculations are for MCS scans, where for N bins we need
    #     N+1 pulses.

    #     Several fields are set in the class:

    #     direction
    #       1 if we are moving positive in user coordinates, −1 if
    #       negative
    #     overall_sense
    #       is our fly motion + or - with respect to encoder counts
    #     taxi_start
    #       The starting point for motor movement during flying, accounts
    #       for needed acceleration of the motor.
    #     taxi_end
    #       The target point for motor movement during flying, accounts
    #       for needed acceleration of the motor.
    #     pso_start
    #       The motor position corresponding to the first PSO pulse.
    #     pso_end
    #        The motor position corresponding to the last PSO pulse.
    #     pso_zero
    #        The motor position where the PSO counter is set to zero.
    #     encoder_step_size
    #       The number of encoder counts for each pixel.
    #     encoder_window_start
    #       The start of the window within which PSO pulses may be emitted,
    #       in encoder counts. Should be slightly wider than the actual PSO
    #       range.
    #     encoder_window_end
    #       The end of the window within which PSO pulses may be emitted,
    #       in encoder counts. Should be slightly wider than the actual PSO
    #     pixel_positions
    #       array of places where pixels are, should occur calculated from
    #       encoder counts then translated to motor positions
    #     """
    #     window_buffer = 5
    #     # Grab any neccessary signals for calculation
    #     start_position = self.flyer_start_position.get()
    #     end_position = self.flyer_end_position.get()
    #     dwell_time = self.flyer_dwell_time.get()
    #     step_size = self.flyer_step_size()
    #     encoder_resolution = self.encoder_resolution.get()
    #     accel_time = self.acceleration.get()
    #     # Check for sane values
    #     if dwell_time == 0:
    #         log.warning(
    #             f"{self} dwell_time is zero. Could not update fly scan parameters."
    #         )
    #         return
    #     if encoder_resolution == 0:
    #         log.warning(
    #             f"{self} encoder resolution is zero. Could not update fly scan"
    #             " parameters."
    #         )
    #         return
    #     if accel_time <= 0:
    #         log.warning(
    #             f"{self} acceleration is non-positive. Could not update fly scan"
    #             " parameters."
    #         )
    #         return
    #     # Determine the desired direction of travel and overall sense
    #     # +1 when moving in + encoder direction, -1 if else
    #     direction = 1 if start_position < end_position else -1
    #     overall_sense = direction * self.encoder_direction
    #     # Calculate the step size in encoder steps
    #     encoder_step_size = round(step_size / encoder_resolution)
    #     # PSO start/end should be located to where req. start/end are
    #     # in between steps. Also doubles as the location where slew
    #     # speed must be met.
    #     pso_start = start_position - (direction * (step_size / 2))
    #     pso_end = end_position + (direction * (step_size / 2))
    #     # Determine taxi distance to accelerate to req speed, v^2/(2*a) = d
    #     # x1.5 for safety margin
    #     slew_speed = step_size / dwell_time
    #     motor_accel = slew_speed / accel_time
    #     taxi_dist = slew_speed**2 / (2 * motor_accel) * 1.5
    #     taxi_start = pso_start - (direction * taxi_dist)
    #     taxi_end = pso_end + (direction * taxi_dist)
    #     # Actually taxi to the first PSO position before the taxi position
    #     encoder_taxi_start = (taxi_start - pso_start) / encoder_resolution
    #     if overall_sense > 0:
    #         rounder = math.floor
    #     else:
    #         rounder = math.ceil
    #     encoder_taxi_start = (
    #         rounder(encoder_taxi_start / encoder_step_size) * encoder_step_size
    #     )
    #     taxi_start = pso_start + encoder_taxi_start * encoder_resolution
    #     # Calculate encoder counts within the requested window of the scan
    #     encoder_window_start = 0  # round(pso_start / encoder_resolution)
    #     encoder_distance = (pso_end - pso_start) / encoder_resolution
    #     encoder_window_end = round(encoder_window_start + encoder_distance)
    #     # Widen the bound a little to make sure we capture the pulse
    #     encoder_window_start -= overall_sense * window_buffer
    #     encoder_window_end += overall_sense * window_buffer

    #     # Check for values outside of the window range for this controller
    #     def is_valid_window(value):
    #         window_in_range = self.encoder_window_min < value < self.encoder_window_max
    #         return window_in_range

    #     window_range = [encoder_window_start, encoder_window_end]
    #     encoder_use_window = all([is_valid_window(v) for v in window_range])
    #     encoder_use_window = encoder_use_window and not self.disable_window.get()
    #     # Create np array of PSO positions in encoder counts
    #     _pso_step = encoder_step_size * overall_sense
    #     _pso_end = encoder_distance + 0.5 * _pso_step
    #     encoder_pso_positions = np.arange(0, _pso_end, _pso_step)
    #     # Transform from PSO positions from encoder counts to engineering units
    #     pso_positions = (encoder_pso_positions * encoder_resolution) + pso_start
    #     # Tranforms from pulse positions to pixel centers
    #     pixel_positions = (pso_positions[1:] + pso_positions[:-1]) / 2
    #     # Set all the calculated variables
    #     [
    #         stat.wait()
    #         for stat in [
    #             self.encoder_step_size.set(encoder_step_size),
    #             self.pso_start.set(pso_start),
    #             self.pso_end.set(pso_end),
    #             self.flyer_slew_speed.set(slew_speed),
    #             self.flyer_taxi_start.set(taxi_start),
    #             self.flyer_taxi_end.set(taxi_end),
    #             self.encoder_window_start.set(encoder_window_start),
    #             self.encoder_window_end.set(encoder_window_end),
    #             self.encoder_use_window.set(encoder_use_window),
    #         ]
    #     ]
    #     self.encoder_pso_positions = encoder_pso_positions
    #     self.pso_positions = pso_positions
    #     self.pixel_positions = pixel_positions

    # def send_command(self, cmd: str):
    #     """Send commands directly to the aerotech ensemble controller.

    #     Returns
    #     =======
    #     status
    #       The Ophyd status object for this write.

    #     """
    #     status = self.parent.asyn.ascii_output.set(cmd, settle_time=0.1)
    #     status.wait()
    #     return status

    # def disable_pso(self):
    #     for axis in range(2):
    #         self.send_command(f"PSOCONTROL @{axis} OFF")

    # def check_flyscan_bounds(self):
    #     """Check that the fly-scan params are sane at the scan start and end.

    #     This checks to make sure no spurious pulses are expected from taxiing.

    #     """
    #     end_points = [
    #         (self.flyer_taxi_start, self.pso_start),
    #         (self.flyer_taxi_end, self.pso_end),
    #     ]
    #     step_size = self.flyer_step_size()
    #     for taxi, pso in end_points:
    #         # Make sure we're not going to have extra pulses
    #         taxi_distance = abs(taxi.get() - pso.get())
    #         if taxi_distance > (1.1 * step_size):
    #             raise InvalidScanParameters(
    #                 f"Scan parameters for {taxi}, {pso}, {self.flyer_step_size} would produce"
    #                 " extra pulses without a window."
    #             )

    # def enable_pso(self):
    #     num_axis = 1
    #     use_window = self.encoder_use_window.get()
    #     if not use_window:
    #         self.check_flyscan_bounds()
    #     # Erase any previous PSO control
    #     self.send_command(f"PSOCONTROL {self.axis} RESET")
    #     # Set the output to occur from the I/O terminal on the
    #     # controller
    #     self.send_command(f"PSOOUTPUT {self.axis} CONTROL {num_axis}")
    #     # Set a pulse 10 us long, 20 us total duration, so 10 us
    #     # on, 10 us off
    #     self.send_command(f"PSOPULSE {self.axis} TIME 20, 10")
    #     # Set the pulses to only occur in a specific window
    #     if use_window:
    #         self.send_command(f"PSOOUTPUT {self.axis} PULSE WINDOW MASK")
    #     else:
    #         self.send_command(f"PSOOUTPUT {self.axis} PULSE")
    #     # Set which encoder we will use.  3 = the MXH (encoder
    #     # multiplier) input. For Ensemble lab, 6 is horizontal encoder
    #     self.send_command(f"PSOTRACK {self.axis} INPUT {self.encoder}")
    #     # Set the distance between pulses in encoder counts
    #     self.send_command(
    #         f"PSODISTANCE {self.axis} FIXED {self.encoder_step_size.get()}"
    #     )
    #     # Which encoder is being used to calculate whether we are
    #     # in the window.
    #     if use_window:
    #         self.send_command(f"PSOWINDOW {self.axis} {num_axis} INPUT {self.encoder}")
    #         # Calculate window function parameters. Must be in encoder
    #         # counts, and is referenced from the stage location where
    #         # we arm the PSO
    #         window_range = (
    #             self.encoder_window_start.get(),
    #             self.encoder_window_end.get(),
    #         )
    #         self.send_command(
    #             f"PSOWINDOW {self.axis} {num_axis} RANGE "
    #             f"{min(window_range)},{max(window_range)}"
    #         )

    # def arm_pso(self):
    #     self.send_command(f"PSOCONTROL {self.axis} ARM")


class AerotechStage(Device):
    """An XY stage for an Aerotech stage with fly-scanning capabilities.

    Parameters
    ==========

    vertical_prefix
      The prefix for the PV of the vertical motor.
    horizontal_prefix
      The prefix to the PV of the horizontal motor.
    delay_prefix
      The prefix to the PVs associated with the pulse delay generator

    """

    _ophyd_labels_ = {"stages"}

    def __init__(
        self,
        vertical_prefix: str,
        horizontal_prefix: str,
        delay_prefix: str | None = None,
        name: str = "",
    ):
        # Axes devices
        self.vert = AerotechMotor(vertical_prefix)
        self.horiz = AerotechMotor(horizontal_prefix)
        # Extra devices for operating the stage
        # asyn = Cpt(AsynRecord, ":asynEns", name="async", labels={"asyns"})
        # A digital delay generator for providing a gate signal
        # delay = FCpt(DG645Delay, "{delay_prefix}:", kind=Kind.config)
        super().__init__(name=name)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from apstools.devices import PTC10AioChannel as PTC10AioChannelBase
from apstools.devices import PTC10PositionerMixin, PTC10TcChannel
from ophyd import Component as Cpt
from ophyd import EpicsSignal, EpicsSignalRO, EpicsSignalWithRBV, PVPositioner

log = logging.getLogger(__name__)


# The apstools version uses "voltage_RBV" as the PVname
class PTC10AioChannel(PTC10AioChannelBase):
    """
    SRS PTC10 AIO module
    """

    voltage = Cpt(EpicsSignalRO, "output_RBV", kind="normal")


class CapillaryHeater(PTC10PositionerMixin, PVPositioner):
    readback = Cpt(EpicsSignalRO, "2A:temperature", kind="hinted")
    setpoint = Cpt(EpicsSignalWithRBV, "5A:setPoint", kind="hinted")

    # Additional modules installed on the PTC10
    pid = Cpt(PTC10AioChannel, "5A:")
    tc = Cpt(PTC10TcChannel, "2A:")
    output_enable = Cpt(EpicsSignal, "outputEnable", kind="omitted")


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
from collections import OrderedDict
from enum import IntEnum
from functools import lru_cache
from typing import Mapping, Sequence

import qtawesome as qta
from bluesky.protocols import HasName, Movable
from ophyd import Device, EpicsMotor, PositionerBase, Signal
from ophyd_async.core import Device as AsyncDevice
from ophyd_async.core import DeviceVector
from ophyd_async.core import Signal as AsyncSignal
from ophyd_async.epics.motor import Motor as EpicsAsyncMotor
from qasync import asyncSlot
from qtpy.QtGui import QColor, QFont, QStandardItem, QStandardItemModel
from qtpy.QtWidgets import (
    QComboBox,
    QHBoxLayout,
    QSizePolicy,
    QToolButton,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from haven.positioner import Positioner

log = logging.getLogger(__name__)


class Flavors(IntEnum):
    UNKNOWN = 0
    VANILLA_DEVICE = 1
    VANILLA_SIGNAL = 2
    ASYNC_DEVICE = 3
    ASYNC_SIGNAL = 4


@lru_cache()
def icons():
    """Produce a dictionary of icons for specific device classes.

    Profiling shows that creating the qta.icon objects is
    slow. Since there are a lot of potential signals, it makes
    sense to cache the icons behind a static method to speed up
    window loading.

    """
    return OrderedDict(
        {
            EpicsMotor: qta.icon("mdi.cog-clockwise"),
            EpicsAsyncMotor: qta.icon("mdi.cog-clockwise"),
            Device: qta.icon("mdi.router-network"),
            AsyncDevice: qta.icon("mdi.router-network"),
            Signal: qta.icon("mdi.connection"),
            AsyncSignal: qta.icon("mdi.connection"),
        }
    )


class DeviceContainer:
    device_flavor: Flavors = Flavors.UNKNOWN

    def _asynchronous_children(self, device):
        for attr_name, child in device.children():
            dot_name = dotted_name(child)
            yield (dot_name, type(child), attr_name)
            yield from self._asynchronous_children(child)

    def _synchronous_children(self, device):
        # Get the subcomponents of the device
        components = getattr(device, "walk_components", lambda: [])()
        # Build the tree from the child components of the device
        for ancestors, dotted_name, cpt in components:
            dotted_name = ".".join([device.name, dotted_name])
            yield (dotted_name, cpt.cls, cpt.attr)


class OphydNode:
    """A node in the Ophyd device hierarchy."""

    device_class: type
    text: str
    dotted_name: str

    def __init__(self, device_class: type, text: str, dotted_name: str):
        self.device_class = device_class
        self.text = text
        self.dotted_name = dotted_name
        # Set the associated
        self.set_items()

    def __str__(self):
        return f"{self.text} ({self.device_class})"

    def set_items(self):
        raise NotImplementedError


class TreeNode(OphydNode):
    parent_item: QStandardItem

    def __init__(self, *args, parent_item, **kwargs):
        self.parent_item = parent_item
        super().__init__(*args, **kwargs)

    def set_items(self):
        """Add model items to the parent item."""
        self.name_item = QStandardItem(self.text)
        parent = self.parent_item
        parent.appendRow(self.name_item)
        row = self.name_item.row()
        # Create an item for the type of component
        type_name = self.device_class.__name__
        self.type_item = QStandardItem(type_name)
        column = 1
        parent.setChild(row, column, self.type_item)
        # Make the component item bold if it's a positioner
        positioner_classes = (PositionerBase, EpicsAsyncMotor, Positioner)
        is_positioner = any(
            issubclass(self.device_class, Klass) for Klass in positioner_classes
        )
        if is_positioner:
            font = QFont()
            font.setBold(True)
            self.name_item.setFont(font)
        # Decide on the row's color
        is_async_device = issubclass(self.device_class, AsyncDevice) and not issubclass(
            self.device_class, AsyncSignal
        )
        is_device = issubclass(self.device_class, Device) or is_async_device
        is_movable = issubclass(self.device_class, Movable)
        if not (is_movable or is_device):
            line_color = QColor("darkgrey")
        else:
            line_color = QColor("black")
        self.name_item.setForeground(line_color)
        self.type_item.setForeground(line_color)
        # Decide on an icon for this component
        for cls, icon in icons().items():
            if issubclass(self.device_class, cls):
                self.type_item.setIcon(icon)
                break
        # Hint non-movable entries
        is_movable = issubclass(self.device_class, Movable)
        if not is_movable:
            font = QFont()
            font.setItalic(True)
            self.name_item.setFont(font)
        # Keep a reference to the component that created the items
        self.name_item.setData(self)
        self.type_item.setData(self)


def dotted_name(obj: HasName) -> str:
    """Get the dotted attribute name of an ophyd_async object."""
    if obj.parent is None:
        # It's a root device, so just the device name
        return obj.name
    # Figure out the attr_name
    if isinstance(obj.parent, DeviceVector):
        attrs = obj.parent
    else:
        attrs = obj.parent.__dict__
    for attr, other_obj in attrs.items():
        if other_obj is obj:
            attr_name = attr
            break
    else:
        print(obj.parent.__class__)
        raise RuntimeError(f"Could not find attribute name for {obj.name}.")
    # siblings = list(attrs.values())
    # attr_names = list(attrs.keys())
    # idx = siblings.index(obj)
    # attr_name = attr_names[siblings.index(obj)]
    # attr_name = list(attrs.keys())[list(attrs.values()).index(obj)]
    # Attach our attr_name to the dotted name of the parent
    parent_name = dotted_name(obj.parent)
    return f"{parent_name}.{attr_name}"


def device_flavor(device):
    # Determine what flavor of device this is
    if isinstance(device, Device):
        return Flavors.VANILLA_DEVICE
    elif isinstance(device, Signal):
        return Flavors.VANILLA_SIGNAL
    elif isinstance(device, AsyncDevice):
        return Flavors.ASYNC_DEVICE
    elif isinstance(device, AsyncSignal):
        return Flavors.ASYNC_SIGNAL
    # Something else, *shrug*
    return Flavors.UNKNOWN


class DeviceTree(DeviceContainer, TreeNode):
    """Representation of an ophyd Component/Device in a tree view."""

    nodes: Mapping
    device_flavor: int = Flavors.UNKNOWN

    def __init__(self, device, *args, **kwargs):
        self.device = device
        self.device_flavor = device_flavor(device)
        # Determine the fully dotted attribute name
        dot_name = dotted_name(device)
        dot_name = kwargs.pop("dotted_name", dot_name)
        super().__init__(
            *args, device_class=type(device), dotted_name=dot_name, **kwargs
        )

    def component_from_dotted_name(self, name):
        return self.nodes[name]

    def add_children(self):
        """Add components of the device as branches on the tree."""
        # Create a place to store the nodes of the tree
        self.nodes = {self.device.name: self}
        # Get the device's children
        if self.device_flavor == Flavors.VANILLA_DEVICE:
            children = self._synchronous_children(self.device)
        elif self.device_flavor == Flavors.ASYNC_DEVICE:
            children = self._asynchronous_children(self.device)
        else:
            children = []
        # Build nodes for the children
        for dotted_name, child_cls, text in children:
            parent = dotted_name.rsplit(".", maxsplit=1)[0]
            parent = self.nodes[parent]
            node = TreeNode(
                parent_item=parent.name_item,
                text=text,
                device_class=child_cls,
                dotted_name=dotted_name,
            )
            self.nodes[dotted_name] = node


class DeviceTreeModel(QStandardItemModel):
    trees: Sequence
    root_item: QStandardItem

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set up the root devices
        self.trees = []
        self.root_item = self.invisibleRootItem()

    def component_from_dotted_name(self, name):
        for tree in self.trees:
            try:
                return tree.component_from_dotted_name(name)
            except KeyError:
                pass
        raise KeyError(name)

    def component_from_index(self, index):
        item = self.itemFromIndex(index)
        return item.data()

    async def add_device(self, device):
        """Add a new root device to the tree."""
        tree = DeviceTree(
            device=device,
            text=device.name,
            parent_item=self.root_item,
        )
        tree.add_children()
        self.trees.append(tree)


class ComboBoxNode(OphydNode):
    def set_items(self):
        """Add model items to the parent item."""
        self.text_item = QStandardItem(self.text)


class DeviceComboBoxModel(DeviceContainer, QStandardItemModel):
    valid_classes = [PositionerBase, Device, AsyncDevice]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.nodes = {}

    async def add_device(self, device):
        """Add components of the device as extra options."""
        # Add a node for the root device itself
        flavor = device_flavor(device)
        root_node = ComboBoxNode(
            text=device.name, device_class=type(device), dotted_name=device.name
        )
        self.nodes[device.name] = root_node
        self.appendRow(root_node.text_item)
        # Get the device's children
        if flavor == Flavors.VANILLA_DEVICE:
            children = self._synchronous_children(device)
        elif flavor == Flavors.ASYNC_DEVICE:
            children = self._asynchronous_children(device)
        else:
            children = []
        # Build the tree from the child components of the device
        # components = getattr(device, "walk_components", lambda: [])()
        # for ancestors, dotted_name, cpt in components:
        # Build nodes for the children
        for dotted_name, child_cls, text in children:
            # Only add a device if it's high-level (e.g. motor)
            is_valid = (issubclass(child_cls, cls) for cls in self.valid_classes)
            if not any(is_valid):
                continue
            # Create the node
            node = ComboBoxNode(
                text=dotted_name, device_class=child_cls, dotted_name=dotted_name
            )
            self.nodes[dotted_name] = node
            # Add the node's model items to the model
            self.appendRow(node.text_item)


class ComponentSelector(QWidget):
    registry = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.create_models()
        self.add_widgets()
        self.connect_signals()

    def current_component(self):
        cpt_name = self.combo_box.currentText()
        if cpt_name == "":
            # Nothing selected
            return None
        else:
            # A device was selected
            return self.registry.find(name=cpt_name)

    def create_models(self):
        self.tree_model = DeviceTreeModel(0, 2)
        self.tree_model.setHorizontalHeaderLabels(["Component", "Type"])
        self.combo_box_model = DeviceComboBoxModel(0, 1)

    def connect_signals(self):
        self.tree_button.toggled.connect(self.tree_view.setVisible)
        self.combo_box.currentTextChanged.connect(self.update_tree_model)
        self.tree_view.selectionModel().currentChanged.connect(
            self.update_combo_box_model
        )

    def update_tree_model(self, new_name):
        log.debug(f"Updating tree: {new_name=}")
        selection = self.tree_view.selectionModel()
        try:
            component = self.tree_model.component_from_dotted_name(new_name)
        except KeyError:
            # It's not a real component, so give up
            log.debug(f"Could not find component for {new_name}, skipping.")
            return
        log.debug(f"Selecting combobox entry: {component.name_item.text()}")
        selection.setCurrentIndex(
            component.name_item.index(), selection.ClearAndSelect | selection.Rows
        )

    def update_combo_box_model(self, index, previous):
        cpt = self.tree_model.component_from_index(index)
        self.combo_box.setCurrentText(cpt.dotted_name)
        log.debug(f"Updating combobox: {cpt.dotted_name=}")

    def select_tree_motor(self):
        if not self.tree_button.isChecked():
            return
        log.debug("Showing motor tree dialog.")
        # Prompt the user to choose a new motor/signal
        accepted = self.tree_dialog.exec()
        self.tree_button.setChecked(False)
        if not accepted:
            # Clicked "Cancel"
            log.debug("Motor tree dialog canceled")
            return
        # Change the current combobox entry
        log.debug("Changing selected motor.")

    @asyncSlot(object)
    async def update_devices(self, registry):
        self.registry = registry
        # Get the devices to add
        devices = sorted(registry.root_devices, key=lambda dev: dev.name.lower())
        for device in devices:
            await self.combo_box_model.add_device(device)
            await self.tree_model.add_device(device)
        # Clear the combobox text so it doesn't auto-select the first entry
        self.combo_box.setCurrentText("")

    def add_widgets(self):
        # Create a layout
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        row_layout = QHBoxLayout()
        layout.addLayout(row_layout)
        # Add a combobox
        combo_box = QComboBox(parent=self)
        combo_box.setModel(self.combo_box_model)
        combo_box.setSizePolicy(
            QSizePolicy.MinimumExpanding,
            QSizePolicy.Fixed,
        )
        row_layout.addWidget(combo_box)
        combo_box.setEditable(True)
        self.combo_box = combo_box
        # Add a button for launching the tree modal dialog
        tree_button = QToolButton(parent=self)
        tree_button.setIcon(qta.icon("mdi.file-tree"))
        tree_button.setCheckable(True)
        row_layout.addWidget(tree_button)
        self.tree_button = tree_button
        # Add a hidden tree view for selecting motors
        tree_view = QTreeView(parent=self)
        tree_view.setVisible(False)
        tree_view.setModel(self.tree_model)
        layout.addWidget(tree_view)
        self.tree_view = tree_view



from pydm.widgets import PyDMByteIndicator
from qtpy.QtWidgets import QFormLayout

from firefly.status import StatusDisplay


def test_shutter_controls(qtbot, shutters, xia_shutter, sim_registry):
    """Do shutter controls get added to the window?"""
    disp = StatusDisplay()
    qtbot.addWidget(disp)
    form = disp.ui.beamline_layout
    # Check label text
    label0 = form.itemAt(4, QFormLayout.LabelRole)
    assert "shutter" in label0.widget().text().lower()
    # Check the widgets for the shutter
    layout0 = form.itemAt(4, QFormLayout.FieldRole)
    indicator = layout0.itemAt(0).widget()
    assert isinstance(indicator, PyDMByteIndicator)
    open_btn = layout0.itemAt(1).widget()
    assert open_btn.text() == "Open"
    close_btn = layout0.itemAt(2).widget()
    assert close_btn.text() == "Close"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from bluesky_queueserver_api import BPlan
from qtpy import QtWidgets
from qtpy.QtGui import QDoubleValidator

from firefly.component_selector import ComponentSelector
from firefly.plans import regions_display

log = logging.getLogger()


class MotorRegion(regions_display.RegionBase):

    def setup_ui(self):
        self.layout = QtWidgets.QHBoxLayout()

        # First item, ComponentSelector
        self.motor_box = ComponentSelector()
        self.layout.addWidget(self.motor_box)

        # Second item, position point
        self.position_line_edit = QtWidgets.QLineEdit()
        self.position_line_edit.setValidator(QDoubleValidator())  # only takes floats
        self.position_line_edit.setPlaceholderText("Position…")
        self.layout.addWidget(self.position_line_edit)


class MoveMotorDisplay(regions_display.RegionsDisplay):
    Region = MotorRegion
    default_num_regions = 1

    def get_scan_parameters(self):
        # get paramters from each rows of line regions:
        motor_lst, position_lst = [], []
        for region_i in self.regions:
            motor_lst.append(region_i.motor_box.current_component().name)
            position_lst.append(float(region_i.position_line_edit.text()))

        motor_args = [
            values for motor_i in zip(motor_lst, position_lst) for values in motor_i
        ]

        return motor_args

    def queue_plan(self, *args, **kwargs):
        """Execute this plan on the queueserver."""
        motor_args = self.get_scan_parameters()

        if self.ui.relative_scan_checkbox.isChecked():
            scan_type = "mvr"
        else:
            scan_type = "mv"

        # # Build the queue item
        item = BPlan(
            scan_type,
            *motor_args,
        )

        # Submit the item to the queueserver
        log.info("Added line scan() plan to queue.")
        self.queue_item_submitted.emit(item)

    def ui_filename(self):
        return "plans/move_motor_window.ui"


# -----------------------------------------------------------------------------
# :author:    Juanjuan Huang
# :email:     juanjuan.huang@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from ophyd_async.core import HintedSignal, Signal

from ..positioner import Positioner
from .monochromator import Monochromator
from .signal import derived_signal_r, derived_signal_rw
from .xray_source import PlanarUndulator

log = logging.getLogger(__name__)


__all__ = ["EnergyPositioner"]


class EnergyPositioner(Positioner):
    """The operational energy of the beamline.

    Responsible for setting both mono and ID energy with an optional
    ID offset. Setting the *energy* component will propagate to both
    real devices and so ``EnergyPositioner().energy`` is a good
    candidate for a positioner for Bluesky plans.

    Currently, the offset between the insertion device and the
    monochromator is fixed. In the future this will be replaced with a
    more sophisticated calculation.

    .. todo::

       Insert functionality to have a non-constant ID offset.

    Attributes
    ==========

    id_offset
      The offset for the insertion device relative to the mono energy.
    energy
      The pseudo positioner for the forward calculation.
    mono_energy
      The real component for the monochromator energy.
    id_energy
      The real component for the insertion device energy.

    Parameters
    ==========
    monochromator_prefix
      The CA prefix for the monochromator IOC.
    undulator_prefix
      The prefix for the insertion device energy, such that
      f"{id_prefix}:Energy.VAL" reaches the energy readback value.

    """

    _ophyd_labels_ = {"energy"}

    def __init__(
        self,
        monochromator_prefix: str,
        undulator_prefix: str,
        name: str = "energy",
    ):
        with self.add_children_as_readables():
            self.monochromator = Monochromator(monochromator_prefix)
            self.undulator = PlanarUndulator(undulator_prefix)
            # Derived positioner signals
            self.setpoint = derived_signal_rw(
                float,
                derived_from={
                    "mono": self.monochromator.energy.user_setpoint,
                    "undulator": self.undulator.energy.setpoint,
                },
                forward=self.set_energy,
                inverse=self.get_energy,
            )
        with self.add_children_as_readables(HintedSignal):
            self.readback = derived_signal_r(
                float,
                derived_from={"mono": self.monochromator.energy.user_readback},
                inverse=self.get_energy,
            )
        # Additional derived signals
        self.precision = derived_signal_rw(
            int, derived_from={"precision": self.monochromator.energy.precision}
        )
        self.units = derived_signal_rw(
            str, derived_from={"units": self.monochromator.energy.motor_egu}
        )
        self.velocity = derived_signal_rw(
            float, derived_from={"velocity": self.monochromator.energy.velocity}
        )

        super().__init__(name=name, put_complete=True)

    async def set_energy(self, value, mono: Signal, undulator: Signal):
        ev_per_kev = 1000
        offset = await self.monochromator.id_offset.get_value()
        vals = {
            self.monochromator.energy: value,
            self.undulator.energy: (value + offset) / ev_per_kev,
        }
        return vals

    def get_energy(self, values, mono: float, undulator: Signal | None = None):
        # Use just the mono value as a readback
        return values[mono]


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.devices.mirrors import HighHeatLoadMirror, KBMirror, KBMirrors


async def test_high_heat_load_mirror_PVs():
    mirror = HighHeatLoadMirror(prefix="255ida:ORM2:", name="orm2", bendable=True)
    await mirror.connect(mock=True)
    # Check the motor PVs
    assert mirror.transverse.user_readback.source == "mock+ca://255ida:ORM2:m1.RBV"
    assert mirror.roll.user_setpoint.source == "mock+ca://255ida:ORM2:m2.VAL"
    assert mirror.upstream.user_setpoint.source == "mock+ca://255ida:ORM2:m3.VAL"
    assert mirror.downstream.user_setpoint.source == "mock+ca://255ida:ORM2:m4.VAL"
    assert mirror.normal.user_setpoint.source == "mock+ca://255ida:ORM2:lateral.VAL"
    assert mirror.pitch.user_setpoint.source == "mock+ca://255ida:ORM2:coarsePitch.VAL"
    assert mirror.bender.user_setpoint.source == "mock+ca://255ida:ORM2:m5.VAL"
    # Check the transform PVs
    assert (
        mirror.drive_transform.channels["B"].input_pv.source
        == "mock+ca://255ida:ORM2:lats:Drive.INPB"
    )
    assert (
        mirror.readback_transform.channels["B"].input_pv.source
        == "mock+ca://255ida:ORM2:lats:Readback.INPB"
    )


async def test_kb_mirrors_PVs():
    kb = KBMirrors(
        prefix="255idcVME:LongKB_Cdn:",
        horiz_upstream_motor="255idcVME:m33",
        horiz_downstream_motor="255idcVME:m34",
        vert_upstream_motor="255idcVME:m35",
        vert_downstream_motor="255idcVME:m36",
        name="kb",
    )
    await kb.connect(mock=True)
    # Check that the individual mirrors are there
    assert isinstance(kb.vert, KBMirror)
    assert isinstance(kb.horiz, KBMirror)
    # Check PVs
    assert (
        kb.horiz.pitch.user_setpoint.source
        == "mock+ca://255idcVME:LongKB_Cdn:H:pitch.VAL"
    )
    assert (
        kb.vert.normal.user_setpoint.source
        == "mock+ca://255idcVME:LongKB_Cdn:V:height.VAL"
    )
    assert kb.horiz.upstream.user_setpoint.source == "mock+ca://255idcVME:m33.VAL"
    assert kb.horiz.downstream.user_setpoint.source == "mock+ca://255idcVME:m34.VAL"
    assert kb.vert.upstream.user_setpoint.source == "mock+ca://255idcVME:m35.VAL"
    assert kb.vert.downstream.user_setpoint.source == "mock+ca://255idcVME:m36.VAL"
    # Check the transforms
    assert (
        kb.horiz.drive_transform.channels["B"].input_pv.source
        == "mock+ca://255idcVME:LongKB_CdnH:Drive.INPB"
    )
    assert (
        kb.horiz.readback_transform.channels["B"].input_pv.source
        == "mock+ca://255idcVME:LongKB_CdnH:Readback.INPB"
    )



import logging

import qtawesome as qta
from pydm.widgets import PyDMByteIndicator, PyDMPushButton
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QHBoxLayout, QSizePolicy

from firefly import display
from haven import beamline
from haven.devices.shutter import ShutterState

log = logging.getLogger(__name__)


def name_to_title(name: str):
    """Convert a python-valid Ophyd object name to a human-readable
    title-case string.

    """
    title = name.replace("_", " ").title()
    return title


class StatusDisplay(display.FireflyDisplay):
    caqtdm_ui_file: str = "/net/s25data/xorApps/ui/25id_main.ui"
    bss_window_requested = Signal()

    def add_shutter_widgets(self):
        form = self.ui.beamline_layout
        # Remove existing layouts
        form.removeRow(self.ui.shutter_A_layout)
        form.removeRow(self.ui.shutter_CD_layout)
        # Add widgets for shutters
        shutters = beamline.registry.findall("shutters", allow_none=True)
        row_idx = 4
        on_color = self.ui.shutter_permit_indicator.onColor
        off_color = self.ui.shutter_permit_indicator.offColor
        for shutter in shutters[::-1]:
            # Add a layout with the buttons
            layout = QHBoxLayout()
            name = shutter.attr_name if shutter.attr_name != "" else shutter.name
            label = name_to_title(name) + ":"
            form.insertRow(row_idx, label, layout)
            # Indicator to show if the shutter is open
            indicator = PyDMByteIndicator(
                parent=self, init_channel=f"haven://{shutter.name}.readback"
            )
            # indicator.showLabels = False
            indicator.labels = ["Closed", "Fault"]
            indicator.numBits = 2
            indicator.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            # Switch colors because open is 0 which should means "good"
            indicator.offColor = on_color
            indicator.onColor = off_color
            layout.addWidget(indicator)
            # Button to open the shutter
            open_btn = PyDMPushButton(
                parent=self,
                label="Open",
                icon=qta.icon("mdi.window-shutter-open"),
                pressValue=ShutterState.OPEN,
                relative=False,
                init_channel=f"haven://{shutter.name}.setpoint",
            )
            open_btn.setEnabled(getattr(shutter, "allow_open", True))
            layout.addWidget(open_btn)
            # Button to close the shutter
            close_btn = PyDMPushButton(
                parent=self,
                label="Close",
                icon=qta.icon("mdi.window-shutter"),
                pressValue=ShutterState.CLOSED,
                relative=False,
                init_channel=f"haven://{shutter.name}.setpoint",
            )
            close_btn.setEnabled(getattr(shutter, "allow_close", True))
            layout.addWidget(close_btn)

    def customize_ui(self):
        self.ui.bss_modify_button.clicked.connect(self.bss_window_requested.emit)
        self.ui.bss_modify_button.setIcon(qta.icon("fa5s.calendar"))
        self.add_shutter_widgets()

    def ui_filename(self):
        return "status.ui"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from bluesky_queueserver_api import BPlan
from pydm import widgets as PyDMWidgets
from pydm.widgets.analog_indicator import PyDMAnalogIndicator
from qtpy import QtWidgets

import haven
from firefly.voltmeters import VoltmetersDisplay
from haven.devices.ion_chamber import IonChamber


@pytest.fixture()
async def ion_chambers(sim_registry):
    devices = []
    for idx, name in enumerate(["I0", "It"]):
        ion_chamber = IonChamber(
            scaler_prefix="255idcVME:3820:",
            scaler_channel=idx,
            preamp_prefix=f"255idc:SR0{idx}",
            voltmeter_prefix="255idc:LJT7_Voltmeter0:",
            voltmeter_channel=idx,
            counts_per_volt_second=10e6,
            name=name,
        )
        await ion_chamber.connect(mock=True)
        sim_registry.register(ion_chamber)
        devices.append(ion_chamber)
    return devices


@pytest.fixture()
async def voltmeters_display(qtbot, ion_chambers, sim_registry):
    vms_display = VoltmetersDisplay()
    qtbot.addWidget(vms_display)
    await vms_display.update_devices(sim_registry)
    return vms_display


@pytest.mark.asyncio
async def test_rows(voltmeters_display):
    """Test that the voltmeters creates a new for each ion chamber."""
    vms_display = voltmeters_display
    # Check that the embedded display widgets get added correctly
    assert hasattr(vms_display, "_ion_chamber_rows")
    num_rows = len(vms_display._ion_chamber_rows)
    assert num_rows == 2
    # two displays and a separator
    expected_count = num_rows * 5
    assert vms_display.voltmeters_layout.count() == expected_count
    # Check that the rows have the correct widgets
    row = vms_display._ion_chamber_rows[0]
    assert isinstance(row.name_label, PyDMWidgets.PyDMLabel)
    assert isinstance(row.voltage_indicator, PyDMAnalogIndicator)
    assert isinstance(row.voltage_label, PyDMWidgets.PyDMLabel)
    assert isinstance(row.voltage_unit_label, QtWidgets.QLabel)
    assert isinstance(row.current_label, PyDMWidgets.PyDMLabel)
    assert isinstance(row.current_unit_label, QtWidgets.QLabel)
    assert isinstance(row.gain_down_button, PyDMWidgets.PyDMPushButton)
    assert isinstance(row.gain_up_button, PyDMWidgets.PyDMPushButton)
    assert isinstance(row.gain_value_label, PyDMWidgets.PyDMLabel)
    assert isinstance(row.gain_unit_label, PyDMWidgets.PyDMLabel)
    assert isinstance(row.auto_gain_checkbox, QtWidgets.QCheckBox)
    assert isinstance(row.details_button, QtWidgets.QPushButton)
    # Check that the widgets are added to the layouts
    assert row.name_label is row.column_layouts[0].itemAt(0).widget()
    assert row.voltage_indicator is row.column_layouts[1].itemAt(0).widget()
    assert row.voltage_label is row.column_layouts[2].itemAt(1).itemAt(1).widget()
    assert row.voltage_unit_label is row.column_layouts[2].itemAt(1).itemAt(2).widget()
    assert row.current_label is row.column_layouts[2].itemAt(2).itemAt(1).widget()
    assert row.current_unit_label is row.column_layouts[2].itemAt(2).itemAt(2).widget()
    assert row.column_layouts[3].itemAt(1).widget().text() == "Gain/Offset"
    assert row.gain_down_button is row.column_layouts[3].itemAt(2).itemAt(1).widget()
    assert row.gain_up_button is row.column_layouts[3].itemAt(2).itemAt(2).widget()
    assert row.gain_value_label is row.column_layouts[3].itemAt(3).itemAt(1).widget()
    assert row.gain_unit_label is row.column_layouts[3].itemAt(3).itemAt(2).widget()
    assert row.auto_gain_checkbox is row.column_layouts[4].itemAt(1).widget()
    # Check that a device has been created properly
    assert isinstance(row.device, haven.IonChamber)


@pytest.mark.asyncio
async def test_gain_button_hints(voltmeters_display, ion_chambers):
    """Test that the gain buttons get disabled when not usable."""
    row = voltmeters_display._ion_chamber_rows[0]
    ic = ion_chambers[0]
    assert row.gain_up_button.isEnabled()
    assert row.gain_down_button.isEnabled()
    # Now set the gain all the way to one limit
    row.update_gain_level_widgets(0)
    assert not row.gain_down_button.isEnabled()
    assert row.gain_up_button.isEnabled()
    # Now set the gain all the way to the other limit
    row.update_gain_level_widgets(27)
    assert row.gain_down_button.isEnabled()
    assert not row.gain_up_button.isEnabled()


def test_details_button(qtbot, voltmeters_display):
    """Check that the details button for each ion chamber triggers the global signal."""
    # Get an embedded display widget
    row = voltmeters_display._ion_chamber_rows[0]
    # Check that the signals are properly connected
    with qtbot.waitSignal(voltmeters_display.ui.details_window_requested, timeout=500):
        row.details_button.click()


@pytest.mark.asyncio
async def test_auto_gain_plan(voltmeters_display, qtbot):
    display = voltmeters_display
    # Put input values into the widgets
    display.ui.volts_min_line_edit.setText("")
    display.ui.volts_max_line_edit.setText("")
    # Check some boxes for which ion chambers get auto-gained
    for idx in [1]:
        row = display._ion_chamber_rows[idx]
        row.auto_gain_checkbox.setChecked(True)
    # Check that the correct plan was sent
    expected_item = BPlan("auto_gain", ["It"])

    def check_item(item):
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        # Simulate clicking on the auto_gain button
        display.ui.auto_gain_button.click()


def test_auto_gain_plan_with_args(qtbot, voltmeters_display):
    display = voltmeters_display
    # Put input values into the widgets
    display.ui.volts_min_line_edit.setText("1.0")
    display.ui.volts_max_line_edit.setText("4.5")
    # Check some boxes for which ion chambers get auto-gained
    for idx in [1]:
        ic_row = display._ion_chamber_rows[idx]
        ic_row.auto_gain_checkbox.setChecked(True)
    # Check that the correct plan was sent
    expected_item = BPlan("auto_gain", ["It"], volts_min=1.0, volts_max=4.5)

    def check_item(item):
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        # Simulate clicking on the auto_gain button
        display.ui.auto_gain_button.click()


@pytest.mark.asyncio
async def test_read_dark_current_plan(voltmeters_display, qtbot):
    display = voltmeters_display
    display.ui.shutter_checkbox.setChecked(True)
    # Check that the correct plan was sent
    expected_item = BPlan(
        "record_dark_current", ["I0", "It"], shutters=["experiment_shutter"]
    )

    def check_item(item):
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        # Simulate clicking on the dark_current button
        # display.ui.dark_current_button.click()
        display.ui.record_dark_current()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



#!/usr/bin/env python3

import logging
from functools import partial

import numpy as np
from caproto import ChannelType
from caproto.server import (
    PVGroup,
    SubGroup,
    get_pv_pair_wrapper,
    ioc_arg_parser,
    pvproperty,
    run,
)
from ophyd.tests.mca_ioc import EpicsDXPGroup, EpicsMCAGroup, MCAROIGroup
from peakutils.peak import gaussian

pvproperty_with_rbv = get_pv_pair_wrapper(setpoint_suffix="", readback_suffix="_RBV")
unknown = int


log = logging.getLogger(__name__)


class DXPGroup(EpicsDXPGroup):
    mca_bin_width = pvproperty(
        value=0.03, name="MCABinWidth_RBV", dtype=float, read_only=True
    )


class ROIGroup(MCAROIGroup):
    # is_hinted = pvproperty(name="BH", dtype=bool)
    hi_chan = pvproperty(name="HI", value=2048)
    lo_chan = pvproperty(name="LO", value=0)
    # hi_chan = pvproperty(name="HI", dtype=int, value=2048)
    # lo_chan = pvproperty(name="LO", dtype=int, value=0)


class MCAGroup(EpicsMCAGroup):
    RoisGroup = type(
        "RoisGroup",
        (PVGroup,),
        {f"roi{i}": SubGroup(ROIGroup, prefix=f".R{i}") for i in range(32)},
    )

    HintsGroup = type(
        "HintsGroup",
        (PVGroup,),
        {
            f"hint{i}": pvproperty(name=f"_R{i}BH", dtype=bool, value="Off")
            for i in range(32)
        },
    )

    rois = SubGroup(RoisGroup, prefix="")
    bluesky_hints = SubGroup(HintsGroup, prefix="")

    erase = pvproperty(name="Erase", dtype=unknown)
    start = pvproperty(name="Start", dtype=unknown)
    preset_real_time = pvproperty(name=".PRTM", value=1.0, dtype=float)
    preset_live_time = pvproperty(name=".PLTM", value=1.0, dtype=float)

    _accumulated_spectrum = None
    spectrum_shape = (2048,)

    @erase.startup
    async def erase(self, instance, async_lib):
        await self.erase.write(1)

    @erase.putter
    async def erase(self, instance, value):
        if value == 1:
            zeros = np.zeros(shape=self.spectrum_shape)
            self._accumulated_spectrum = zeros
            # await self.spectrum.write(self._accumulated_spectrum)

    @start.startup
    async def start(self, instance, async_lib):
        self._async_lib = async_lib

    @start.putter
    async def start(self, instance, value):
        # Wait for the preset time to expire
        await self._async_lib.library.sleep(self.preset_real_time.value)
        # Set a simulated spectrum
        spectrum = self._generate_spectrum()
        await self.spectrum.write(spectrum)

    def _generate_spectrum(self):
        bin_width = self.parent.dxp1.mca_bin_width.value
        shape = self.spectrum_shape
        energies = np.arange(*shape) * bin_width
        # Empty spectrum
        spectrum = np.zeros(shape)
        # Add peaks
        peaks = [
            (8.047, 100, 0.2),  # Cu Ka1
            (8.905, 50, 0.2),  # Cu Kb
            (22.163, 150, 0.5),  # Ag Ka
            (24.942, 75, 0.5),  # Ag Kb
            (49.128, 20, 0.7),  # Er Ka
            (55.681, 10, 0.7),  # Er Kb
        ]
        for energy, height, sigma in peaks:
            height *= np.random.rand() * 0.2 + 0.9
            energy *= np.random.rand() * 0.05 + 0.975
            sigma *= np.random.rand() * 0.2 + 0.9
            spectrum += gaussian(energies, height, energy, sigma)
        # Add noise
        spectrum += 3 * np.random.rand(*shape)
        # Add it to the list of accumulated spectra
        self._accumulated_spectrum += spectrum
        return self._accumulated_spectrum

    spectrum = pvproperty(name="", dtype=float, read_only=True, value=np.zeros((2048,)))


class VortexME4IOC(PVGroup):
    async def propogate_to_mcas(self, instance, value, field):
        """Share this value with all the elements in the detector."""
        # print(f"Propogating {value} to {field} (async_lib: {self._async_lib}.")
        mcas = [self.mca1, self.mca2, self.mca3, self.mca4]
        coroutines = [getattr(mca, field).write(value) for mca in mcas]
        await self._async_lib.library.gather(*coroutines)

    mca1 = SubGroup(MCAGroup, prefix="mca1")
    mca2 = SubGroup(MCAGroup, prefix="mca2")
    mca3 = SubGroup(MCAGroup, prefix="mca3")
    mca4 = SubGroup(MCAGroup, prefix="mca4")
    dxp1 = SubGroup(DXPGroup, prefix="dxp1:")
    dxp2 = SubGroup(DXPGroup, prefix="dxp2:")
    dxp3 = SubGroup(DXPGroup, prefix="dxp3:")
    dxp4 = SubGroup(DXPGroup, prefix="dxp4:")

    start_all = pvproperty(
        name="StartAll",
        dtype=unknown,
        # value="Done",
        # record="mbbi",
        # enum_strings=("Done", "Start"),
        # dtype=ChannelType.ENUM,
    )
    erase_start = pvproperty(name="EraseStart", dtype=unknown)  # value="Done",
    # record="mbbi",
    # enum_strings=("Done", "Start"),
    # dtype=ChannelType.ENUM)
    erase_all = pvproperty(name="EraseAll", dtype=unknown)
    stop_all = pvproperty(name="StopAll", dtype=unknown)
    acquiring = pvproperty(
        name="Acquiring",
        record="mbbi",
        value="Done",
        enum_strings=("Done", "Acquiring"),
        dtype=ChannelType.ENUM,
    )
    preset_real_time = pvproperty(
        name="PresetReal",
        value=1.0,
        dtype=float,
        put=partial(propogate_to_mcas, field="preset_real_time"),
    )
    preset_live_time = pvproperty(
        name="PresetLive",
        value=1.0,
        dtype=float,
        put=partial(propogate_to_mcas, field="preset_live_time"),
    )
    dead_time = pvproperty(name="DeadTime", dtype=float, read_only=True)

    @start_all.startup
    async def start_all(self, instance, async_lib):
        self._async_lib = async_lib

    @start_all.putter
    async def start_all(self, instance, value):
        """Start all the elements in the detector."""
        log.debug(f"Received start_all value: {value}")
        if value in ("Done", 0):
            # No-op for setting "Done"
            return
        await self.acquiring.write("Acquiring")
        await self.propogate_to_mcas(instance, value, field="start")
        # await self._async_lib.library.sleep(self.preset_real_time.value)
        dead_time = np.random.rand() * 30 + 5
        await self.dead_time.write(dead_time)
        await self.acquiring.write("Done")

    @erase_start.putter
    async def erase_start(self, instance, value):
        log.debug(f"Erase-starting: {value}")
        await self.erase_all.write(value)
        await self.start_all.write(value)

    @erase_all.putter
    async def erase_all(self, instance, value):
        log.debug(f"Erasing all: {value}")
        await self.propogate_to_mcas(instance, value, field="erase")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="vortex_me4:", desc="ophyd.tests.test_mca test IOC"
    )
    ioc = VortexME4IOC(**ioc_options)
    run(ioc.pvdb, **run_options)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import time
import warnings
from collections import OrderedDict
from enum import IntEnum
from typing import Optional, Sequence

from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as DDC
from ophyd import Kind, Signal, flyers, mca
from ophyd.signal import DerivedSignal, InternalSignal
from ophyd.status import StatusBase, SubscriptionStatus

from .. import exceptions
from .fluorescence_detector import (
    MCASumMixin,
    ROIMixin,
    UseROISignal,
    XRFMixin,
    active_kind,
    add_roi_sums,
)

__all__ = ["DxpDetector"]


NUM_ROIS = 32


class SizeSignal(DerivedSignal):
    def inverse(self, hi):
        lo = self.parent.lo_chan.get()
        size = hi - lo
        return size

    def forward(self, size):
        lo = self.parent.lo_chan.get()
        hi = lo + size
        return hi


class ROI(ROIMixin, mca.ROI):
    _default_read_attrs = [
        "count",
        "net_count",
    ]
    _default_configuration_attrs = [
        "label",
        "bkgnd_chans",
        "hi_chan",
        "lo_chan",
    ]
    kind = active_kind
    # Signals
    use = Cpt(UseROISignal, derived_from="label", kind="config")
    size = Cpt(SizeSignal, derived_from="hi_chan", kind="config")

    def unstage(self):
        # Restore original signal kinds
        for fld, kind in self._original_kinds.items():
            getattr(self, fld).kind = kind
        super().unstage()


def add_rois(range_: Sequence[int] = range(NUM_ROIS), kind=Kind.normal, **kwargs):
    """Add one or more ROIs to an MCA instance

    Parameters
    ----------
    range_ : sequence of ints
        Must be be in the set [0,31]

    By default, an EpicsMCA is initialized with all 32 rois. These
    provide the following Components as EpicsSignals (N=[0,31]):
    EpicsMCA.rois.roiN.(label,count,net_count,preset_cnt, is_preset,
    bkgnd_chans, hi_chan, lo_chan)

    """
    defn = OrderedDict()
    kwargs["kind"] = kind
    for roi in range_:
        if not (0 <= roi < 32):
            raise ValueError("roi must be in the set [0,31]")
        attr = "roi{}".format(roi)
        defn[attr] = (
            ROI,
            ".R{}".format(roi),
            kwargs,
        )
    return defn


class MCARecord(MCASumMixin, mca.EpicsMCARecord):
    rois = DDC(add_rois(), kind=active_kind)
    dead_time_factor = Cpt(Signal, kind=Kind.normal)
    dead_time_percent = Cpt(Signal, kind=Kind.normal)
    _default_read_attrs = [
        "rois",
        "total_count",
        "spectrum",
        "dead_time_factor",
        "dead_time_percent",
        # "preset_real_time",
        # "preset_live_time",
        # "elapsed_real_time",
        # "elapsed_live_time",
        # "background",
    ]
    _default_configuration_attrs = ["rois", "mode"]
    kind = active_kind


def add_mcas(range_, kind=active_kind, **kwargs):
    """Add one or more MCARecords to a device

    Parameters
    ----------
    range_
      Indices for which to create MCA records.

    """
    defn = OrderedDict()
    kwargs["kind"] = kind
    for idx in range_:
        attr = f"mca{idx}"
        defn[attr] = (
            MCARecord,
            f"mca{idx+1}",  # (Epics uses 1-index instead of 0-index)
            kwargs,
        )
    return defn


class DxpDetector(
    XRFMixin,
    flyers.FlyerInterface,
    mca.EpicsDXPMapping,
    mca.EpicsDXPMultiElementSystem,
):
    """A fluorescence detector based on XIA-DXP XMAP electronics.

    Creates MCA components based on the number of elements.

    """

    class CollectMode(IntEnum):
        MCA_SPECTRA = 0
        MCA_MAPPING = 1
        SCA_MAPPING = 2
        LIST_MAPPING = 3

    write_path: str = "M:\\epics\\fly_scanning\\"
    read_path: str = "/net/s20data/sector20/tmp/"
    acquire = Cpt(Signal, value=0)

    # Dead time aggregate statistics
    dead_time_average = Cpt(InternalSignal, kind="normal")
    dead_time_min = Cpt(InternalSignal, kind="normal")
    dead_time_max = Cpt(InternalSignal, kind="normal")

    # By default, a 4-element detector, subclass for more elements
    mcas = DDC(
        add_mcas(range_=range(4)),
        default_read_attrs=["mca0", "mca1", "mca2", "mca3"],
        default_configuration_attrs=["mca0", "mca1", "mca2", "mca3"],
    )
    roi_sums = DDC(
        add_roi_sums(mcas=range(1), rois=range(NUM_ROIS)),
        kind=active_kind,
        default_read_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
        default_configuration_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
    )

    # net_cdf = Cpt(NetCDFPlugin_V34, "netCDF1:")
    _default_read_attrs = [
        # "preset_live_time",
        # "preset_real_time",
        # "dead_time",
        # "elapsed_live",
        # "elapsed_real",
        "roi_sums",
        "mcas",
        "dead_time_average",
        "dead_time_max",
        "dead_time_min",
    ]
    _default_configuration_attrs = [
        "max_scas",
        "num_scas",
        "poll_time",
        "prescale",
        "preset_mode",
        "preset_events",
        "preset_triggers",
        "snl_connected",
        "mcas",
    ]
    _omitted_attrs = [  # Just save them here for easy reference
        "channel_advance",
        "client_wait",
        "dwell",
        "save_system",
        "save_system_file",
        "set_client_wait",
        "erase_all",
        "erase_start",
        "start_all",
        "stop_all",
        "set_acquire_busy",
        "acquire_busy",
        "status_all",
        "status_all_once",
        "acquiring",
        "read_baseline_histograms",
        "read_all",
        "read_all_once",
        "copy_adcp_ercent_rule",
        "copy_baseline_cut_enable",
        "copy_baseline_cut_percent",
        "copy_baseline_filter_length",
        "copy_baseline_threshold",
        "copy_decay_time",
        "copy_detector_polarity",
        "copy_energy_threshold",
        "copy_gap_time",
        "copy_max_energy",
        "copy_max_width",
        "copy_peaking_time",
        "copy_preamp_gain",
        "copy_roic_hannel",
        "copy_roie_nergy",
        "copy_roi_sca",
        "copy_reset_delay",
        "copy_trigger_gap_time",
        "copy_trigger_peaking_time",
        "copy_trigger_threshold",
        "do_read_all",
        "do_read_baseline_histograms",
        "do_read_traces",
        "do_status_all",
        "read_low_level_params",
        "read_traces",
        "trace_modes",
        "trace_times",
        "idead_time",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs[self.collect_mode] = self.CollectMode.MCA_SPECTRA
        # Listen for changes to the ``acquire`` channel.
        self.acquire.subscribe(self._acquire)
        self.acquiring.subscribe(self._acquire)

    @property
    def default_time_signal(self):
        return self.preset_real_time

    def _acquire(self, *args, old_value, value, obj, **kwargs):
        """Mimic the Xspress3 AD interface for acquiring data."""
        if obj is self.acquire:
            # Update the real signals
            if bool(value):
                # Start
                self.erase_start.set(1).wait()
            else:
                # Stop
                self.stop_all.set(1).wait()
        elif obj is self.acquiring:
            # Update the virtual signal
            self.acquire.set(value).wait()

    def rois(self, roi_indices: Optional[Sequence[int]] = None):
        # Get the list of ROIs to activate
        all_rois = []
        for mca in self.mca_records():
            rois = mca.rois.component_names
            rois = [getattr(mca.rois, r) for r in rois]
            # Get sub-list if requested
            if roi_indices is not None:
                rois = [rois[i] for i in roi_indices]
            all_rois.extend(rois)
        return all_rois

    def kickoff(self) -> StatusBase:
        """Start the detector flying.

        This starts acquisition, starts the file writer, and sets the
        detector to advance to the next when it receives a gate
        signal. The returned status object will be complete when the
        detector reports that it is acquiring.

        Returns
        =======
        StatusBase
          Becomes complete once the detector is acquiring and ready to
          measure data.

        """
        # Make sure the CDF file write plugin is primed (assumes
        # dimensions will be empty when not primed)
        is_primed = len(self.net_cdf.dimensions.get()) > 0
        if not is_primed:
            msg = f"{self.net_cdf.name} plugin not primed."
            warnings.warn(msg, RuntimeWarning)
            exc = exceptions.PluginNotPrimed(msg)
            status = StatusBase()
            status.set_exception(exc)
            return status

        # Set up the status for when the detector is ready to fly
        def check_acquiring(*, old_value, value, **kwargs):
            is_acquiring = bool(value)
            if is_acquiring:
                self.start_timestamp = time.time()
            return is_acquiring

        status = SubscriptionStatus(self.acquiring, check_acquiring)
        # Configure the mapping controls
        self.collect_mode.set(self.CollectMode.MCA_MAPPING)
        self.pixel_advance_mode.set("Gate")
        # Configure the netCDF file writer
        self.net_cdf.enable.set("Enable").wait()
        [
            status.wait()
            for status in [
                self.net_cdf.file_path.set(self.write_path),
                self.net_cdf.file_name.set("fly_scan_temp.nc"),
                self.net_cdf.file_write_mode.set("Capture"),
            ]
        ]
        self.net_cdf.capture.set(1).wait()
        # Start the detector
        self.erase_start.set(1)
        return status

    def complete(self) -> StatusBase:
        """Wait for the detector to finish flying.

        This will stop the file writers and acquisition then return a
        status object. The status object will report complete once the
        detector acquire status reports that the detector is not
        acquiring.

        Returns
        =======
        StatusBase
          Becomes complete once the detector has finished acquiring
          and is ready to report the collected data.

        """
        # Stop the CDF file writer
        self.net_cdf.capture.set(0).wait()
        self.stop_all.set(1)

        # Set up the status for when the detector is done collecting
        def check_acquiring(*, old_value, value, **kwargs):
            is_acquiring = bool(value)
            return not is_acquiring

        status = SubscriptionStatus(self.acquiring, check_acquiring)
        return status


def parse_xmap_buffer(buff):
    """Extract meaningful data from an XMAP internal buffer during mapping.

    For more information, see section 5.3.3 of
    https://cars9.uchicago.edu/software/epics/XMAP_User_Manual.pdf

    """
    data = {"header": {}}
    header = buff[:256]
    # Verify tag words
    assert header[0] == 0x55AA
    assert header[1] == 0xAA55
    # Parse remaining buffer header
    head_data = data["header"]
    head_data["buffer_header_size"] = header[2]
    head_data["mapping_mode"] = header[3]
    head_data["run_number"] = header[4]
    head_data["buffer_number"] = header[5:7]
    head_data["buffer_id"] = header[7]
    head_data["num_pixels"] = header[8]
    head_data["starting_pixel"] = header[9:11]
    head_data["module"] = header[11]
    head_data["buffer_overrun"] = header[24]
    # head_data[""] =
    return data


def make_dxp_device(name, prefix, num_elements):
    # Build the mca components
    mca_range = range(0, num_elements)
    attrs = {
        "mcas": DDC(
            add_mcas(range_=mca_range),
            kind=active_kind,
            default_read_attrs=[f"mca{i}" for i in mca_range],
            default_configuration_attrs=[f"mca{i}" for i in mca_range],
        ),
        "roi_sums": DDC(
            add_roi_sums(mcas=mca_range, rois=range(NUM_ROIS)),
            kind=active_kind,
            default_read_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
            default_configuration_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
        ),
    }
    # Create a dynamic subclass with the MCAs
    class_name = name.title().replace("_", "")
    parent_classes = (DxpDetector,)
    Cls = type(name, parent_classes, attrs)
    return Cls(
        prefix=prefix,
        name=name,
        labels={"xrf_detectors", "fluorescence_detectors", "detectors"},
    )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from PyQt5.QtGui import QStandardItem, QStandardItemModel
from qtpy.QtWidgets import QAbstractItemView, QListView


class DetectorListView(QListView):
    detector_model: QStandardItemModel

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Create the data model
        self.detector_model = QStandardItemModel()
        self.setModel(self.detector_model)
        # Make it possible to select multiple detectors
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)

    async def update_devices(self, registry):
        # Get devices
        detectors = registry.findall(label="detectors", allow_none=True)
        # Remove old detectors list from model
        model = self.detector_model
        model.removeRows(0, model.rowCount())
        # Add new detectors to model
        for det in detectors:
            model.appendRow(QStandardItem(det.name))

    def selected_detectors(self):
        indexes = self.selectedIndexes()
        items = [self.detector_model.itemFromIndex(i) for i in indexes]
        names = [item.text() for item in items]
        return names


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest.mock import AsyncMock, MagicMock

import pytest
from ophyd import Device
from ophyd.sim import make_fake_device
from ophydregistry import Registry

import firefly
from firefly.action import WindowAction
from firefly.controller import FireflyController
from firefly.queue_client import QueueClient


@pytest.fixture()
async def controller(qapp):
    controller = FireflyController()
    await controller.setup_instrument(load_instrument=False)
    return controller


@pytest.fixture()
def ffapp():
    return MagicMock()


def test_prepare_queue_client(controller):
    api = MagicMock()
    controller.prepare_queue_client(api=api)
    assert isinstance(controller._queue_client, QueueClient)


def test_queue_actions_enabled(controller, qtbot):
    """Check that the queue control bottons only allow sensible actions.

    For example, if the queue is idle, the "abort" button should be
    disabled, among many others.

    """
    actions = controller.actions.queue_controls
    # Pretend the queue has some things in it
    with qtbot.waitSignal(controller.queue_re_state_changed, timeout=1000):
        controller.queue_re_state_changed.emit("idle")
    # Check the enabled state of all the buttons
    assert actions["start"].isEnabled()
    assert not actions["pause"].isEnabled()
    assert not actions["pause_now"].isEnabled()
    assert not actions["stop_runengine"].isEnabled()
    assert not actions["resume"].isEnabled()
    assert not actions["abort"].isEnabled()
    assert not actions["stop_queue"].isEnabled()
    assert not actions["halt"].isEnabled()
    # Pretend the queue has been paused
    with qtbot.waitSignal(controller.queue_re_state_changed, timeout=1000):
        controller.queue_re_state_changed.emit("paused")
    # Check the enabled state of all the buttons
    assert not actions["start"].isEnabled()
    assert not actions["pause"].isEnabled()
    assert not actions["pause_now"].isEnabled()
    assert not actions["stop_queue"].isEnabled()
    assert actions["stop_runengine"].isEnabled()
    assert actions["resume"].isEnabled()
    assert actions["abort"].isEnabled()
    assert actions["halt"].isEnabled()
    # Pretend the queue is running
    with qtbot.waitSignal(controller.queue_re_state_changed, timeout=1000):
        controller.queue_re_state_changed.emit("running")
    # Check the enabled state of all the buttons
    assert not actions["start"].isEnabled()
    assert actions["pause"].isEnabled()
    assert actions["pause_now"].isEnabled()
    assert not actions["stop_runengine"].isEnabled()
    assert actions["stop_queue"].isEnabled()
    assert not actions["resume"].isEnabled()
    assert not actions["abort"].isEnabled()
    assert not actions["halt"].isEnabled()
    # Pretend the queue is in an unknown state (maybe the environment is closed)
    with qtbot.waitSignal(controller.queue_re_state_changed, timeout=1000):
        controller.queue_re_state_changed.emit(None)


@pytest.fixture()
def tardis(sim_registry):
    Tardis = make_fake_device(Device)
    tardis = Tardis(name="my_tardis", labels={"tardis"})
    sim_registry.register(tardis)
    return tardis


def test_prepare_generic_device_windows(controller, tardis, mocker):
    """Check for preparing devices with the ``show_device_window`` slot."""
    # mocker.patch.object(controller, "show_device_window", autospec=True)
    actions = controller.device_actions(device_label="tardis", display_file="tardis.ui")
    # Check that actions were created
    assert "my_tardis" in actions.keys()
    assert isinstance(actions["my_tardis"], WindowAction)


@pytest.mark.asyncio
async def test_load_instrument_registry(controller, qtbot, monkeypatch):
    """Check that the instrument registry gets created."""
    assert isinstance(controller.registry, Registry)
    # Mock the underlying haven instrument loader
    loader = AsyncMock()
    monkeypatch.setattr(firefly.controller.beamline, "load", loader)
    monkeypatch.setattr(controller, "prepare_queue_client", MagicMock())
    # Reload the devices and see if the registry is changed
    with qtbot.waitSignal(controller.registry_changed):
        await controller.setup_instrument(load_instrument=True)
    # Make sure we loaded the instrument
    assert loader.called


###########################################################
# Tests for connecting the queue client and the controller
###########################################################


def test_queue_stopped(controller):
    """Does the action respond to changes in the queue stopped pending?"""
    client = controller.prepare_queue_client(api=MagicMock())
    assert not controller.actions.queue_controls["stop_queue"].isChecked()
    client.queue_stop_changed.emit(True)
    assert controller.actions.queue_controls["stop_queue"].isChecked()
    client.queue_stop_changed.emit(False)
    assert not controller.actions.queue_controls["stop_queue"].isChecked()


def test_autostart_changed(controller, qtbot):
    """Does the action respond to changes in the queue autostart
    status?

    """
    client = controller.prepare_queue_client(api=MagicMock())
    autostart_action = controller.actions.queue_settings["autostart"]
    autostart_action.setChecked(True)
    assert autostart_action.isChecked()
    with qtbot.waitSignal(client.autostart_changed, timeout=3):
        client.autostart_changed.emit(False)
    assert not autostart_action.isChecked()
    with qtbot.waitSignal(client.autostart_changed, timeout=3):
        client.autostart_changed.emit(True)
    assert autostart_action.isChecked()


@pytest.mark.asyncio
async def test_ion_chamber_details_window(qtbot, sim_registry, ion_chamber, controller):
    """Check that the controller opens ion chamber windows from voltmeters
    display.

    """
    vm_action = controller.actions.voltmeter
    ic_action = controller.actions.ion_chambers[ion_chamber.name]
    # Create the ion chamber display
    vm_action = controller.actions.voltmeter
    with qtbot.waitSignal(vm_action.window_shown, timeout=1):
        vm_action.trigger()
    vm_display = vm_action.window.display_widget()
    qtbot.addWidget(vm_display)
    await vm_display.update_devices(sim_registry)
    # Click the ion chamber button
    details_button = vm_display._ion_chamber_rows[0].details_button
    with qtbot.waitSignal(ic_action.window_shown, timeout=1000):
        details_button.click()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import warnings
from collections import OrderedDict
from typing import Dict, Generator

import numpy as np
from ophyd import Component as Cpt
from ophyd import Device, Kind, Signal, get_cl
from ophyd.flyers import FlyerInterface
from ophyd.status import Status
from scipy.interpolate import CubicSpline

log = logging.getLogger()


warnings.warn(
    DeprecationWarning(
        "Motor flyer will be removed in a future release. Please replace with the ophyd-async Motor for fly-scan support"
    )
)


class MotorFlyer(FlyerInterface, Device):
    # Desired fly parameters
    flyer_start_position = Cpt(Signal, name="start_position", value=0, kind=Kind.config)
    flyer_end_position = Cpt(Signal, name="end_position", value=1, kind=Kind.config)
    # step_size = Cpt(Signal, name="step_size", value=1, kind=Kind.config)
    flyer_num_points = Cpt(Signal, value=2, kind=Kind.config)
    flyer_dwell_time = Cpt(Signal, value=1, kind=Kind.config)

    # Calculated fly parameters
    flyer_slew_speed = Cpt(Signal, value=1, kind=Kind.config)
    flyer_taxi_start = Cpt(Signal, kind=Kind.config)
    flyer_taxi_end = Cpt(Signal, kind=Kind.config)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._kickoff_thread = None
        self._complete_thread = None
        self._fly_model = None
        self.cl = get_cl()
        # Set up auto-calculations for the flyer
        self.motor_egu.subscribe(self._update_fly_params)
        self.acceleration.subscribe(self._update_fly_params)
        self.flyer_start_position.subscribe(self._update_fly_params)
        self.flyer_end_position.subscribe(self._update_fly_params)
        self.flyer_num_points.subscribe(self._update_fly_params)
        self.flyer_dwell_time.subscribe(self._update_fly_params)

    def kickoff(self):
        """Start the motor as a flyer.

        The status object return is marked as done once flying
        is ready.

        Returns
        -------
        kickoff_status : StatusBase
            Indicate when flying is ready.

        """
        self.log.debug(f"Kicking off {self}")

        def kickoff_thread():
            try:
                self.move(self.flyer_taxi_start.get(), wait=True)
                self.velocity.put(self.flyer_slew_speed.get())
            except Exception as exc:
                st.set_exception(exc)
            else:
                self.log.debug(f"{self} kickoff succeeded")
                st.set_finished()
            finally:
                # keep a local reference to avoid any GC shenanigans
                th = self._kickoff_thread
                # these two must be in this order to avoid a race condition
                self._kickoff_thread = None
                del th

        if self._kickoff_thread is not None:
            raise RuntimeError(
                "Another kickoff() call is still in progress " f"for {self.name}"
            )

        st = Status(self)
        self._status = st
        self._kickoff_thread = self.cl.thread_class(target=kickoff_thread)
        self._kickoff_thread.daemon = True
        self._kickoff_thread.start()
        return self._status

    def complete(self):
        """Start the motor flying and wait for it to complete.

        Returns
        -------
        complete_status : StatusBase
            Indicate when flying has completed

        """
        self.log.debug(f"Comleting {self}")

        def complete_thread():
            try:
                # Record real motor positions for later evaluation
                self._fly_data = []
                self._fly_model = None
                cid = self.user_readback.subscribe(self.record_datum, run=False)
                self.move(self.flyer_taxi_end.get(), wait=True)
                self.user_readback.unsubscribe(cid)
            except Exception as exc:
                st.set_exception(exc)
            else:
                self.log.debug(f"{self} complete succeeded")
                st.set_finished()
            finally:
                # keep a local reference to avoid any GC shenanigans
                th = self._complete_thread
                # these two must be in this order to avoid a race condition
                self._complete_thread = None
                del th

        if self._complete_thread is not None:
            raise RuntimeError(
                f"Another complete() call is still in progress for {self.name}"
            )

        st = Status(self)
        self._status = st
        self._complete_thread = self.cl.thread_class(target=complete_thread)
        self._complete_thread.daemon = True
        self._complete_thread.start()
        return self._status

    def record_datum(self, *, old_value, value, timestamp, **kwargs):
        """Record a fly-scan data point so we can report it later."""
        self._fly_data.append((timestamp, value))

    def collect(self) -> Generator[Dict, None, None]:
        """Retrieve data from the flyer as proto-events

        Yields
        ------
        event_data : dict
            Must have the keys {'time', 'timestamps', 'data'}.

        """
        # Create the data objects
        for time, position in self._fly_data:
            yield {
                "time": time,
                "timestamps": {
                    self.user_readback.name: time,
                    self.user_setpoint.name: time,
                },
                "data": {
                    self.user_readback.name: position,
                    self.user_setpoint.name: position,
                },
            }

    def predict(self, timestamp: float) -> Dict:
        """Predict where the motor was at *timestamp* during most recent fly
        scan.

        Parameters
        ==========
        timestamp
          The unix timestamp to use for interpolating the measured data.

        Returns
        =======
        datum
          A data event for this timestamp similar to those provided by
          the ``collect()`` method.

        """
        # Prepare an interpolation model for fly scan data
        if self._fly_model is None:
            times, positions = np.asarray(self._fly_data).transpose()
            self._fly_model = CubicSpline(times, positions, bc_type="clamped")
        model = self._fly_model
        # Interpolate the data value based on timestamp
        position = float(model(timestamp))
        setpoint = self.pixel_positions[
            np.argmin(np.abs(self.pixel_positions - position))
        ]
        datum = {
            "time": timestamp,
            "timestamps": {
                self.user_readback.name: timestamp,
                self.user_setpoint.name: timestamp,
            },
            "data": {
                self.user_readback.name: position,
                self.user_setpoint.name: setpoint,
            },
        }
        return datum

    def describe_collect(self):
        """Describe details for the collect() method"""
        desc = OrderedDict()
        desc.update(self.describe())
        return {self.user_readback.name: desc}

    def _update_fly_params(self, *args, **kwargs):
        """Calculate new fly-scan parameters based on signal values.

        Computes several parameters describing the fly scan motion.
        These include the actual start position of the motor, the
        actual distance between points, and the end position of the
        motor.

        Several fields are set in the class:

        direction
          1 if we are moving positive in user coordinates, −1 if
          negative
        taxi_start
          The starting point for motor movement during flying, accounts
          for needed acceleration of the motor.
        taxi_end
          The target point for motor movement during flying, accounts
          for needed acceleration of the motor.
        pixel_positions
          array of places where pixels are, should occur calculated from
          encoder counts then translated to motor positions

        """
        # Grab any neccessary signals for calculation
        start_position = self.flyer_start_position.get()
        end_position = self.flyer_end_position.get()
        dwell_time = self.flyer_dwell_time.get()
        num_points = self.flyer_num_points.get()
        accel_time = self.acceleration.get()
        # Check for sane values
        if dwell_time == 0:
            log.warning(
                f"{self} dwell_time is zero. Could not update fly scan parameters."
            )
            return
        if accel_time <= 0:
            log.warning(
                f"{self} acceleration is non-positive. Could not update fly scan"
                " parameters."
            )
            return
        # Determine the desired direction of travel:
        # +1 when moving in + encoder direction, -1 if else
        direction = 1 if start_position < end_position else -1
        # Determine taxi distance to accelerate to req speed, v^2/(2*a) = d
        # x1.5 for safety margin
        step_size = abs((start_position - end_position) / (num_points - 1))
        if step_size <= 0:
            log.warning(
                f"{self} step_size is non-positive. Could not update fly scan"
                " parameters."
            )
            return
        slew_speed = step_size / dwell_time
        motor_accel = slew_speed / accel_time
        taxi_dist = slew_speed**2 / (2 * motor_accel) * 1.5 + step_size / 2
        taxi_start = start_position - (direction * taxi_dist)
        taxi_end = end_position + (direction * taxi_dist)
        # Tranforms from pulse positions to pixel centers
        pixel_positions = np.linspace(start_position, end_position, num=num_points)
        # Set all the calculated variables
        self.flyer_slew_speed.put(slew_speed)
        self.flyer_taxi_start.put(taxi_start)
        self.flyer_taxi_end.put(taxi_end)
        self.pixel_positions = pixel_positions



"""A QPushButton that responds to the state of the queue server."""

import logging

import qtawesome as qta
from qtpy import QtGui, QtWidgets
from strenum import StrEnum

log = logging.getLogger(__name__)


class Colors(StrEnum):
    ADD_TO_QUEUE = "rgb(0, 123, 255)"
    RUN_QUEUE = "rgb(25, 135, 84)"


class QueueButton(QtWidgets.QPushButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Initially disable the button until the status of the queue can be determined
        self.setDisabled(True)

    def update_queue_style(self, status: dict):
        if status["worker_environment_exists"]:
            self.setEnabled(True)
        else:
            # Should be disabled because the queue is closed
            self.setDisabled(True)
        # Coloration for the whether the item would get run immediately
        if status["re_state"] == "idle" and status["queue_autostart_enabled"]:
            # Will play immediately
            self.setStyleSheet(
                f"background-color: {Colors.RUN_QUEUE};\n"
                f"border-color: {Colors.RUN_QUEUE};"
            )
            self.setIcon(qta.icon("fa5s.play"))
            if self.text() in ["Run", "Add to Queue", ""]:
                self.setText("Run")
            self.setToolTip("Add this plan to the queue and start it immediately.")
        elif status["worker_environment_exists"]:
            # Will be added to the queue
            self.setStyleSheet(
                f"background-color: {Colors.ADD_TO_QUEUE};\n"
                f"border-color: {Colors.ADD_TO_QUEUE};\n"
            )
            self.setIcon(qta.icon("fa5s.list"))
            if self.text() in ["Run", "Add to Queue", ""]:
                self.setText("Add to Queue")
            self.setToolTip("Add this plan to the queue to run later.")
        else:
            # Regular old (probably disabled) button
            self.setStyleSheet("")
            self.setIcon(QtGui.QIcon())


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from bluesky import plan_stubs as bps

from ..instrument import beamline
from ..motor_position import rbv

log = logging.getLogger(__name__)


__all__ = ["robot_transfer_sample"]


ON = 1


def robot_transfer_sample(robot, sampleN, *args):
    """
    Use robot to load sampleN at a fixed Aerotech stage position or any
    motors.

    e.g.
    Load sample:
    robot_sample(robot, 9, motor1, 100, motor2, 200, motor3, 50)

    Unload sample:
    robot_sample(robot, None, motor1, 100, motor2, 200, motor3, 50)

    Parameters
    ==========
    robot
      robot device
    sampleN
      Sample number
    args
      multiple pairs of motor and pos
    """

    robot = beamline.registry.find(robot)

    # Check if power is on
    # if robot.power_rbv.get() == Off:
    #    raise ValueError("Robot is NOT powered. Please turn on the power.")

    # Check if remote control is on
    # if not robot.remote_control.get() == Off:
    #    raise ValueError(
    #        "Robot is NOT in reomete control mode. Please click on Pad for remote control."
    #    )

    # Check if gripper is on
    # if robot.gripper_activated.get() == Off:
    #    raise ValueError("Gripper is not activated. Please activate the gripper.")

    # Check if robot is running
    # if robot.grogram_running.get() == On:
    #   raise ValueError("Robot is running now.")

    # Find motors
    motor_list = [beamline.registry.find(motor) for motor in args[::2]]
    new_positions = [pos for pos in args[1::2]]
    # Record the motor positions before load sampls
    initial_positions = [rbv(motor) for motor in motor_list]

    # Move the aerotech to the loading position
    for motor, pos in zip(motor_list, new_positions):
        yield from bps.mv(motor, pos)

    if sampleN == None:
        # Unload sample
        yield from bps.mv(robot.unload_current_sample, ON)

    else:
        # Load sampleN after all the other motor arrive at (pos1, pos2, pos3...)
        sample = getattr(
            robot.samples, f"sample{sampleN}"
        )  # Access the Sample device corresponding to sampleN
        print(sample.load)
        yield from bps.mv(sample.load, ON)  # Assuming '1' initiates the loading action

    # Return to the initial position
    for motor, pos in zip(motor_list[-1::-1], initial_positions[-1::-1]):
        yield from bps.mv(motor, pos)


# -----------------------------------------------------------------------------
# :author:    Yanna Chen
# :email:     yannachen@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

from bluesky_queueserver_api import BPlan

from firefly import display

log = logging.getLogger()


class CountDisplay(display.FireflyDisplay):

    async def update_devices(self, registry):
        """Set available components in the device list."""
        await super().update_devices(registry)
        await self.ui.detectors_list.update_devices(registry)

    def customize_ui(self):
        self.ui.run_button.clicked.connect(self.queue_plan)

    def queue_plan(self, *args, **kwargs):
        """Execute this plan on the queueserver."""
        # Get scan parameters from widgets
        num_readings = self.ui.num_spinbox.value()
        delay = self.ui.delay_spinbox.value()
        detectors = self.ui.detectors_list.selected_detectors()
        # Build the queue item

        item = BPlan("count", delay=delay, num=num_readings, detectors=detectors)
        # Submit the item to the queueserver
        log.info("Add ``count()`` plan to queue.")
        self.queue_item_submitted.emit(item)

    def ui_filename(self):
        return "plans/count.ui"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import time
import warnings
from collections import OrderedDict, namedtuple
from enum import IntEnum
from typing import Dict, Mapping

import numpy as np
import pandas as pd
from apstools.devices import CamMixin_V34, SingleTrigger_V34
from ophyd import ADComponent as ADCpt
from ophyd import CamBase
from ophyd import Component as Cpt
from ophyd import DetectorBase as OphydDetectorBase
from ophyd import (
    Device,
    EigerDetectorCam,
    EpicsSignal,
    Kind,
    Lambda750kCam,
    OphydObject,
    Signal,
    SimDetectorCam,
    SingleTrigger,
)
from ophyd.areadetector.base import EpicsSignalWithRBV as SignalWithRBV
from ophyd.areadetector.filestore_mixins import (
    FileStoreHDF5IterativeWrite,
    FileStoreTIFFIterativeWrite,
)
from ophyd.areadetector.plugins import (
    HDF5Plugin_V31,
    HDF5Plugin_V34,
    ImagePlugin_V31,
    ImagePlugin_V34,
    OverlayPlugin,
    OverlayPlugin_V34,
    PvaPlugin_V31,
    PvaPlugin_V34,
    ROIPlugin_V31,
    ROIPlugin_V34,
)
from ophyd.areadetector.plugins import StatsPlugin_V31 as OphydStatsPlugin_V31
from ophyd.areadetector.plugins import StatsPlugin_V34 as OphydStatsPlugin_V34
from ophyd.areadetector.plugins import TIFFPlugin_V31, TIFFPlugin_V34
from ophyd.flyers import FlyerInterface
from ophyd.sim import make_fake_device
from ophyd.status import Status, StatusBase, SubscriptionStatus

from .. import exceptions
from .._iconfig import load_config

log = logging.getLogger(__name__)


__all__ = ["Eiger500K", "Lambda250K", "SimDetector", "AsyncCamMixin"]


class WriteModes(IntEnum):
    SINGLE = 0
    CAPTURE = 1
    STREAM = 2


class Capture(IntEnum):
    STOP = 0
    START = 1


class ImageMode(IntEnum):
    SINGLE = 0
    MULTIPLE = 1
    CONTINUOUS = 2


class EraseState(IntEnum):
    DONE = 0
    ERASE = 1


class AcquireState(IntEnum):
    DONE = 0
    ACQUIRE = 1


class TriggerMode(IntEnum):
    SOFTWARE = 0
    INTERNAL = 1
    IDC = 2
    TTL_VETO_ONLY = 3
    TTL_BOTH = 4
    LVDS_VETO_ONLY = 5
    LVDS_BOTH = 6


class DetectorState(IntEnum):
    IDLE = 0
    ACQUIRE = 1
    READOUT = 2
    CORRECT = 3
    SAVING = 4
    ABORTING = 5
    ERROR = 6
    WAITING = 7
    INITIALIZING = 8
    DISCONNECTED = 9
    ABORTED = 10


fly_event = namedtuple("fly_event", ("timestamp", "value"))


class FlyerMixin(FlyerInterface, Device):
    flyer_num_points = Cpt(Signal)
    flyscan_trigger_mode = TriggerMode.SOFTWARE

    def save_fly_datum(self, *, value, timestamp, obj, **kwargs):
        """Callback to save data from a signal during fly-scanning."""
        datum = fly_event(timestamp=timestamp, value=value)
        self._fly_data.setdefault(obj, []).append(datum)

    def kickoff(self) -> StatusBase:
        # Set up subscriptions for capturing data
        self._fly_data = {}
        for walk in self.walk_fly_signals():
            sig = walk.item
            # Run subs the first time to make sure all signals are present
            sig.subscribe(self.save_fly_datum, run=True)

        # Set up the status for when the detector is ready to fly
        def check_acquiring(*, old_value, value, **kwargs):
            is_acquiring = value == DetectorState.ACQUIRE
            if is_acquiring:
                self.start_fly_timestamp = time.time()
            return is_acquiring

        status = SubscriptionStatus(self.cam.detector_state, check_acquiring)
        # Set the right parameters
        self._original_vals.setdefault(self.cam.image_mode, self.cam.image_mode.get())
        status &= self.cam.image_mode.set(ImageMode.CONTINUOUS)
        status &= self.cam.trigger_mode.set(self.flyscan_trigger_mode)
        status &= self.cam.num_images.set(2**14)
        status &= self.cam.acquire.set(AcquireState.ACQUIRE)
        return status

    def complete(self) -> StatusBase:
        """Wait for flying to be complete.

        This commands the Xspress to stop acquiring fly-scan data.

        Returns
        -------
        complete_status : StatusBase
          Indicate when flying has completed
        """
        # Remove subscriptions for capturing fly-scan data
        for walk in self.walk_fly_signals():
            sig = walk.item
            sig.clear_sub(self.save_fly_datum)
        self.cam.acquire.set(AcquireState.DONE)
        return Status(done=True, success=True, settle_time=0.5)

    def collect(self) -> dict:
        """Generate the data events that were collected during the fly scan."""
        # Load the collected data, and get rid of extras
        fly_data, fly_ts = self.fly_data()
        fly_data.drop("timestamps", inplace=True, axis="columns")
        fly_ts.drop("timestamps", inplace=True, axis="columns")
        # Yield each row one at a time
        for data_row, ts_row in zip(fly_data.iterrows(), fly_ts.iterrows()):
            payload = {
                "data": {sig.name: val for (sig, val) in data_row[1].items()},
                "timestamps": {sig.name: val for (sig, val) in ts_row[1].items()},
                "time": float(np.median(np.unique(ts_row[1].values))),
            }
            yield payload

    def describe_collect(self) -> Dict[str, Dict]:
        """Describe details for the flyer collect() method"""
        desc = OrderedDict()
        # for walk in self.walk_fly_signals():
        #     desc.update(walk.item.describe())
        return {self.name: self.describe()}

    def fly_data(self):
        """Compile the fly-scan data into a pandas dataframe."""
        # Get the data for frame number as a reference
        image_counter = pd.DataFrame(
            self._fly_data[self.cam.array_counter],
            columns=["timestamps", "image_counter"],
        )
        image_counter["image_counter"] -= 2  # Correct for stray frames
        # Build all the individual signals' dataframes
        dfs = []
        for sig, data in self._fly_data.items():
            df = pd.DataFrame(data, columns=["timestamps", sig])
            old_shape = df.shape
            nums = (df.timestamps - image_counter.timestamps).abs()

            # Assign each datum an image number based on timestamp
            def get_image_num(ts):
                """Get the image number taken closest to a given timestamp."""
                num = image_counter.iloc[
                    (image_counter["timestamps"] - ts).abs().argsort()[:1]
                ]
                num = num["image_counter"].iloc[0]
                return num

            im_nums = [get_image_num(ts) for ts in df.timestamps.values]
            df.index = im_nums
            # Remove duplicates and intermediate ROI sums
            df.sort_values("timestamps")
            df = df.groupby(df.index).last()
            dfs.append(df)
        # Combine frames into monolithic dataframes
        data = image_counter.copy()
        data = data.set_index("image_counter", drop=True)
        timestamps = data.copy()
        for df in dfs:
            sig = df.columns[1]
            data[sig] = df[sig]
            timestamps[sig] = df["timestamps"]
        # Fill in missing values, most likely because the value didn't
        # change so no new camonitor reply was received
        data = data.ffill(axis=0)
        timestamps = timestamps.ffill(axis=1)
        # Drop the first frame since it was just the result of all the subs
        data.drop(data.index[0], inplace=True)
        timestamps.drop(timestamps.index[0], inplace=True)
        return data, timestamps

    def walk_fly_signals(self, *, include_lazy=False):
        """Walk all signals in the Device hierarchy that are to be read during
        fly-scanning.

        Parameters
        ----------
        include_lazy : bool, optional
            Include not-yet-instantiated lazy signals

        Yields
        ------
        ComponentWalk
            Where ancestors is all ancestors of the signal, including the
            top-level device `walk_signals` was called on.

        """
        for walk in self.walk_signals():
            # Image counter has to be included for data alignment
            if walk.item is self.cam.array_counter:
                yield walk
                continue
            # Only include readable signals
            if not bool(walk.item.kind & Kind.normal):
                continue
            # ROI sums do not get captured properly during flying
            # Instead, they should be calculated at the end
            # if self.roi_sums in walk.ancestors:
            #     continue
            yield walk


class AsyncCamMixin(OphydObject):
    """A mixin that allows for delayed evaluation of the connection status.

    Ordinarily, and area detector gets the camera acquire signal,
    which requires that it be connected. With this mixin, you can skip
    that step, so that you can ``wait_for_connection()`` fully at a
    later time.

    """

    lazy_wait_for_connection = False

    # Components
    acquire = ADCpt(SignalWithRBV, "Acquire")


class SingleImageModeTrigger(SingleTrigger_V34):
    """A trigger mixin for cameras that don't support "Multiple" image mode."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "cam.image_mode" in self.stage_sigs:
            self.stage_sigs["cam.image_mode"] = ImageMode.SINGLE


class SimDetectorCam_V34(CamMixin_V34, SimDetectorCam): ...


class EigerCam(AsyncCamMixin, EigerDetectorCam): ...


class LambdaCam(AsyncCamMixin, Lambda750kCam): ...


class StageCapture:
    """Mixin to prepare NDPlugin file capture mode.

    Sets the number of captures to zero (infinite), and starts
    capturing. Then when the device gets unstaged, capturing turns
    back off.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Stage the capture button as well
        self.stage_sigs[self.file_write_mode] = WriteModes.STREAM
        self.stage_sigs[self.capture] = Capture.START
        self.stage_sigs[self.num_capture] = 0


class DynamicFileStore(Device):
    """File store mixin that alters the write_path_template based on
    iconfig values.

    """

    def __init__(
        self, *args, write_path_template="/{root_path}/%Y/%m/{name}/", **kwargs
    ):
        super().__init__(*args, write_path_template=write_path_template, **kwargs)
        # Format the file_write_template with per-device values
        config = load_config()
        try:
            self.write_path_template = self.write_path_template.format(
                name=self.parent.name,
                root_path=config.get("area_detector_root_path", "tmp"),
            )
        except KeyError:
            warnings.warn(f"Could not format write_path_template {write_path_template}")

    def _add_dtype_str(self, desc: Mapping) -> Mapping:
        """Add the specific image data type into the metadata.

        This method modifies the dictionary in place.

        Parameters
        ==========
        desc:
          The input description, most likely coming from self.describe().

        Returns
        =======
        desc
          The same dictionary, with an added ``dtype_str`` key.

        """
        key = f"{self.parent.name}_image"
        if key in desc:
            dtype = self.data_type.get(as_string=True)
            dtype_str = np.dtype(dtype.lower()).str
            desc[key].setdefault("dtype_str", dtype_str)
        return desc

    def describe(self):
        return self._add_dtype_str(super().describe())


class HDF5FilePlugin(DynamicFileStore, FileStoreHDF5IterativeWrite, HDF5Plugin_V34):
    """
    Add data acquisition methods to HDF5Plugin.
    * ``stage()`` - prepare device PVs befor data acquisition
    * ``unstage()`` - restore device PVs after data acquisition
    * ``generate_datum()`` - coordinate image storage metadata
    """

    def stage(self):
        self.stage_sigs.move_to_end("capture", last=True)
        super().stage()


class TIFFFilePlugin(DynamicFileStore, FileStoreTIFFIterativeWrite, TIFFPlugin_V34): ...


class DetectorBase(FlyerMixin, OphydDetectorBase):
    def __init__(self, *args, description=None, **kwargs):
        super().__init__(*args, **kwargs)
        if description is None:
            description = self.name
        self.description = description

    @property
    def default_time_signal(self):
        return self.cam.acquire_time


class StatsMixin:
    overlays = ADCpt(OverlayPlugin, "Over1:")
    _default_read_attrs = [
        "max_value",
        "min_value",
        "min_xy.x",
        "max_xy.x",
        "min_xy.y",
        "max_xy.y",
        "total",
        "net",
        "mean_value",
        "sigma_value",
    ]


class StatsPlugin_V31(StatsMixin, OphydStatsPlugin_V31): ...


class StatsPlugin_V34(StatsMixin, OphydStatsPlugin_V34): ...


class SimDetector(SingleTrigger_V34, DetectorBase):
    """
    ADSimDetector
    SingleTrigger:
    * stop any current acquisition
    * sets image_mode to 'Multiple'
    """

    cam = ADCpt(SimDetectorCam_V34, "cam1:")
    image = ADCpt(ImagePlugin_V34, "image1:")
    pva = ADCpt(PvaPlugin_V34, "Pva1:")
    hdf1 = ADCpt(
        HDF5FilePlugin,
        "HDF1:",
        write_path_template="/tmp/",
    )
    roi1 = ADCpt(ROIPlugin_V34, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V34, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V34, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V34, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V34, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V34, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V34, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V34, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V34, "Stats5:", kind=Kind.normal)
    overlays = ADCpt(OverlayPlugin_V34, "Over1:")


class TIFFPlugin(StageCapture, TIFFPlugin_V31):
    _default_read_attrs = ["full_file_name"]


class HDF5Plugin(StageCapture, HDF5Plugin_V31):
    _default_read_attrs = ["full_file_name"]


class Lambda250K(SingleTrigger, DetectorBase):
    """
    A Lambda 250K area detector device.
    """

    cam = ADCpt(LambdaCam, "cam1:")
    image = ADCpt(ImagePlugin_V31, "image1:")
    pva = ADCpt(PvaPlugin_V31, "Pva1:")
    tiff = ADCpt(TIFFPlugin, "TIFF1:", kind=Kind.normal)
    hdf1 = ADCpt(HDF5Plugin, "HDF1:", kind=Kind.normal)
    roi1 = ADCpt(ROIPlugin_V31, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V31, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V31, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V31, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V31, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V31, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V31, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V31, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V31, "Stats5:", kind=Kind.normal)

    _default_read_attrs = [
        "stats1",
        "stats2",
        "stats3",
        "stats4",
        "stats5",
        "hdf1",
        "tiff",
    ]


class Eiger500K(SingleTrigger, DetectorBase):
    """
    A Eiger S 500K area detector device.
    """

    cam = ADCpt(EigerCam, "cam1:")
    image = ADCpt(ImagePlugin_V34, "image1:")
    pva = ADCpt(PvaPlugin_V34, "Pva1:")
    tiff = ADCpt(TIFFPlugin, "TIFF1:", kind=Kind.normal)
    hdf1 = ADCpt(HDF5Plugin, "HDF1:", kind=Kind.normal)
    roi1 = ADCpt(ROIPlugin_V34, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V34, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V34, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V34, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V34, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V34, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V34, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V34, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V34, "Stats5:", kind=Kind.normal)

    _default_read_attrs = [
        "stats1",
        "stats2",
        "stats3",
        "stats4",
        "stats5",
        "hdf1",
        "tiff",
    ]


class AravisCam(AsyncCamMixin, CamBase):
    gain_auto = ADCpt(EpicsSignal, "GainAuto")
    acquire_time_auto = ADCpt(EpicsSignal, "ExposureAuto")


class AravisDetector(SingleImageModeTrigger, DetectorBase):
    """
    A gige-vision camera described by EPICS.
    """

    _default_configuration_attrs = (
        "cam",
        "hdf",
        "stats1",
        "stats2",
        "stats3",
        "stats4",
    )
    _default_read_attrs = ("cam", "hdf", "stats1", "stats2", "stats3", "stats4")

    cam = ADCpt(AravisCam, "cam1:")
    image = ADCpt(ImagePlugin_V34, "image1:")
    pva = ADCpt(PvaPlugin_V34, "Pva1:")
    overlays = ADCpt(OverlayPlugin_V34, "Over1:")
    roi1 = ADCpt(ROIPlugin_V34, "ROI1:", kind=Kind.config)
    roi2 = ADCpt(ROIPlugin_V34, "ROI2:", kind=Kind.config)
    roi3 = ADCpt(ROIPlugin_V34, "ROI3:", kind=Kind.config)
    roi4 = ADCpt(ROIPlugin_V34, "ROI4:", kind=Kind.config)
    stats1 = ADCpt(StatsPlugin_V34, "Stats1:", kind=Kind.normal)
    stats2 = ADCpt(StatsPlugin_V34, "Stats2:", kind=Kind.normal)
    stats3 = ADCpt(StatsPlugin_V34, "Stats3:", kind=Kind.normal)
    stats4 = ADCpt(StatsPlugin_V34, "Stats4:", kind=Kind.normal)
    stats5 = ADCpt(StatsPlugin_V34, "Stats5:", kind=Kind.normal)
    hdf = ADCpt(HDF5FilePlugin, "HDF1:", kind=Kind.normal)
    # tiff = ADCpt(TIFFFilePlugin, "TIFF1:", kind=Kind.normal)


def make_area_detector(prefix: str, name: str, device_class: str, mock=True) -> Device:
    # Create the area detectors defined in the configuration
    try:
        DeviceClass = globals().get(device_class)
    except TypeError:
        msg = f"area_detector.{name}.device_class={device_class}"
        raise exceptions.UnknownDeviceConfiguration(msg)
    # Create a simulated version if needed
    if mock:
        DeviceClass = make_fake_device(DeviceClass)
    # Create the device co-routine
    device = DeviceClass(
        prefix=prefix,
        name=name,
        labels={"area_detectors", "detectors"},
    )
    return device


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
